from collections.abc import Iterable
from typing import Any, Dict, List, Optional, Union

import albumentations as A
from omegaconf import DictConfig, ListConfig
from torch import Tensor
from torch.nn import Module, Parameter
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from torchok.constructor import DATASETS, LOSSES, OPTIMIZERS, SCHEDULERS, TRANSFORMS
from torchok.constructor.config_structure import Phase
from torchok.data.datasets.base import ImageDataset
from torchok.losses.base import JointLoss
from torchok.metrics.metrics_manager import MetricsManager


class Constructor:
    """Provides factory features for optimizers, schedulers, loss functions, data loaders and metrics."""

    def __init__(self, hparams: DictConfig):
        """Init Constructor with hparams.

        Args:
            hparams: Configuration dictionary for all the mentioned components.
            The dictionary should contain at least the following parameter groups (see configuration for details):
            - optimization
            - data
            - losses
            - metrics
        """
        self.__hparams = hparams

    def configure_optimizers(self, parameters: Union[Module, Tensor, List[Union[Module, Tensor]]],
                             optim_idx: int = -1) -> List[Dict[str, Union[Optimizer, Dict[str, Any]]]]:
        """Create optimizers and learning rate schedulers from a pre-defined configuration.

        Note: optimization parameters are split into two groups: decay and no-decay specifying which parameters can be
        weight decayed and which not. Each module is checked on having an attribute `no_weight_decay`,
        specifying a list of submodules which must be not weight decayed (useful in transformer models such as Swin).
        All *.bias parameters, 1D tensors and scalars are put into no-decay group according to the best practice.

        Args:
            parameters: Parameters for optimization
            optim_idx: Optimizer and scheduler index if specific optimizer/scheduler group is needed to be created.
            Default is -1 meaning all the optimizers/schedulers are created.

        Returns:
            List of dicts in the PyTorch Lightning accepted format:
            - optimizer: PyTorch-like optimizer
            - lr_scheduler: dict
                scheduler: PyTorch-like scheduler
                <any PyTorch Lightning parameters. See `their documentation`_>

        Raises:
            - ValueError: When a requested optimizer/scheduler group with optim_idx isn't present in configuration
            - ValueError: When parameters type is out of supported types list (see typing)

        .. _their documentation:
        https://pytorch-lightning.readthedocs.io/en/1.6.2/common/lightning_module.html#configure-optimizers
        """
        optims_params = self.__hparams.optimization
        if 0 <= optim_idx < len(optims_params):
            optims_params = [optims_params[optim_idx]]
        elif optim_idx >= len(optims_params):
            raise ValueError(f'You requested optimization with index {optim_idx} while '
                             f'there\'re only {len(optims_params)} optimization parameters are specified')

        opt_sched_list = []
        for optim_params in optims_params:
            optimizer = self.__create_optimizer(parameters, optim_params.optimizer)
            opt_sched = {'optimizer': optimizer}

            if optim_params.scheduler is not None:
                scheduler_dict = self.__create_scheduler(optimizer, optim_params.scheduler)
                opt_sched['lr_scheduler'] = scheduler_dict

            opt_sched_list.append(opt_sched)

        return opt_sched_list

    @staticmethod
    def __create_optimizer(parameters: Union[Module, Tensor, List[Union[Module, Tensor]]],
                           optimizer_params: DictConfig) -> Optimizer:
        optimizer_class = OPTIMIZERS.get(optimizer_params.name)
        parameters = Constructor.__set_weight_decay_for_parameters(parameters)
        optimizer = optimizer_class(parameters, **optimizer_params.params)

        return optimizer

    @staticmethod
    def __create_scheduler(optimizer: Optimizer, scheduler_params: DictConfig) -> Dict[str, Any]:
        scheduler_class = SCHEDULERS.get(scheduler_params.name)
        scheduler = scheduler_class(optimizer, **scheduler_params.params)
        pl_params = scheduler_params.pl_params

        return {
            'scheduler': scheduler,
            **pl_params
        }

    @staticmethod
    def __set_weight_decay_for_parameters(parameters: Union[Module, Tensor, List[Union[Module, Tensor]]]) -> List[
            Dict[str, Union[Tensor, float]]]:
        if not isinstance(parameters, Iterable) and not isinstance(parameters, Module) and \
           not isinstance(parameters, Tensor):
            raise ValueError(f'Unsupported parameters type for optimizer: {type(parameters)}')
        elif not isinstance(parameters, Iterable):
            parameters = [parameters]

        param_groups = []
        for model in parameters:
            if isinstance(model, Module):
                param_groups.extend(Constructor.__param_groups_weight_decay(model))
            elif isinstance(model, Tensor) and model.requires_grad:
                param_groups.append({'params': model})

        return param_groups

    # Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/optim/optim_factory.py
    # Copyright 2019 Ross Wightman
    # Licensed under The Apache 2.0 License [see LICENSE for details]
    @staticmethod
    def __param_groups_weight_decay(model: Module) -> List[Dict[str, Union[Parameter, float]]]:
        # Module names for which weights decay will not be used.
        no_weight_decay_list = []
        
        if hasattr(model, 'no_weight_decay'):
            no_weight_decay_list = model.no_weight_decay()

        decay = []
        no_decay = []
        for name, param in model.named_parameters():
            if not param.requires_grad:
                continue

            if param.ndim <= 1 or name.endswith('.bias') or name in no_weight_decay_list:
                no_decay.append(param)
            else:
                decay.append(param)

        return [
            {'params': no_decay, 'weight_decay': 0.},
            {'params': decay}]

    def create_dataloaders(self, phase: Phase) -> List[DataLoader]:
        """Create data loaders.

        Each data loader is based on a dataset while dataset consists
        augmentations and transformations in the `albumentations`_ format.

        Args:
            phase: Phase for which the data loaders are to be built.

        Returns:
            List of data loaders to be used inside `PyTorch Lightning Module`_
        .. _albumentations: https://albumentations.ai/docs
        .. _PyTorch Lightning Module:
        https://pytorch-lightning.readthedocs.io/en/1.6.2/common/lightning_module.html#train-dataloader

        Raises:
            - ValueError: When transforms are not specified for composition augmentations of albumentation
            - ValueError: When OneOrOther composition is passed that isn't supported
        """
        if phase in self.hparams.data:
            return [
                self.__prepare_dataloader(phase_params.dataset, phase_params.dataloader)
                for phase_params in self.hparams.data[phase] if phase_params is not None
            ]
        else:
            return []

    @staticmethod
    def __prepare_dataloader(dataset_params: DictConfig, dataloader_params: DictConfig) -> DataLoader:
        dataset = Constructor._create_dataset(dataset_params)
        collate_fn = dataset.collate_fn if hasattr(dataset, 'collate_fn') else None

        loader = DataLoader(dataset=dataset,
                            collate_fn=collate_fn,
                            **dataloader_params)

        return loader

    @staticmethod
    def _create_dataset(dataset_params: DictConfig) -> ImageDataset:
        transform = Constructor.__create_transforms(dataset_params.transform)
        # TODO remove when OmegaConf is fixing the bug, write to issue to Omegaconf!
        # Config structure had 'augment' parameter with default value = None, but in loaded config 
        # 'augment' is not in keys of dataset_params dictionary. So it must be written like
        # augment_params = dataset_params.augment
        augment_params = dataset_params.get('augment', None)
        augment = Constructor.__create_transforms(augment_params)

        dataset_class = DATASETS.get(dataset_params.name)

        return dataset_class(transform=transform, augment=augment, **dataset_params.params)

    @staticmethod
    def __prepare_transforms_recursively(transforms: ListConfig) -> List[Union[A.Compose, A.BaseCompose]]:
        transforms_list = []

        for transform_info in transforms:
            transform_name = transform_info.name
            transform_params = transform_info.get('params', dict())

            if transform_name == 'Compose' or transform_name == 'OneOf' or transform_name == 'SomeOf' or \
               transform_name == 'PerChannel' or transform_name == 'Sequential':
                transform = Constructor.__prepare_base_compose(transform_name, **transform_params)
            elif transform_name == 'OneOrOther':
                raise ValueError('OneOrOther composition is currently not supported')
            else:
                transform = TRANSFORMS.get(transform_name)(**transform_params)

            transforms_list.append(transform)

        return transforms_list

    @staticmethod
    def __prepare_base_compose(compose_name: str, **kwargs) -> A.Compose:
        transforms = kwargs.pop('transforms', None)
        if transforms is None:
            raise ValueError(f'There are transforms must be specified for {compose_name} composition')

        transforms_list = Constructor.__prepare_transforms_recursively(transforms)
        transform = TRANSFORMS.get(compose_name)(transforms=transforms_list, **kwargs)

        return transform

    @staticmethod
    def __create_transforms(transforms_params: ListConfig) -> Optional[A.Compose]:
        if transforms_params is None:
            return None

        return Constructor.__prepare_base_compose('Compose', transforms=transforms_params)

    def configure_metrics_manager(self):
        """Create list of metrics wrapping them into a MetricManager module.

        Returns: MetricManager module.
        """
        return MetricsManager(self.__hparams.metrics)

    def configure_losses(self) -> JointLoss:
        """Create list of loss modules wrapping them into a JointLoss module.

        Returns: JointLoss module
        """
        loss_modules, mappings, tags, weights = [], [], [], []
        for loss_config in self.__hparams.joint_loss.losses:
            loss_module = LOSSES.get(loss_config.name)(**loss_config.params)
            loss_modules.append(loss_module)
            mappings.append(loss_config.mapping)
            tags.append(loss_config.tag)
            weights.append(loss_config.weight)

        normalize_weights = self.__hparams.joint_loss.normalize_weights

        return JointLoss(loss_modules, mappings, tags, weights, normalize_weights)

    @property
    def hparams(self) -> DictConfig:
        """Return configuration dictionary.

        Returns:
            Dict, containing at least the following parameter groups (see configuration for details):
                - optimization
                - data
                - losses
                - metrics
        """
        return self.__hparams



from typing import Dict, Union

import torch
from omegaconf import DictConfig

from torchok.constructor import BACKBONES, HEADS, NECKS, POOLINGS, TASKS
from torchok.tasks.base import BaseTask


@TASKS.register_class
class ClassificationTask(BaseTask):
    """A class for image classification task."""

    def __init__(self, hparams: DictConfig):
        """Init ClassificationTask.

        Args:
            hparams: Hyperparameters that set in yaml file.
        """
        super().__init__(hparams)
        # BACKBONE
        backbone_name = self._hparams.task.params.get('backbone_name')
        backbones_params = self._hparams.task.params.get('backbone_params', dict())
        self.backbone = BACKBONES.get(backbone_name)(**backbones_params)

        # NECK
        neck_name = self._hparams.task.params.get('neck_name')
        neck_params = self._hparams.task.params.get('neck_params', dict())
        neck_in_channels = self.backbone.out_channels

        if neck_name is not None:
            self.neck = NECKS.get(neck_name)(in_channels=neck_in_channels, **neck_params)
            pooling_in_channels = self.neck.out_channels
        else:
            self.neck = None
            pooling_in_channels = neck_in_channels

        # POOLING
        pooling_params = self._hparams.task.params.get('pooling_params', dict())
        pooling_name = self._hparams.task.params.get('pooling_name')
        self.pooling = POOLINGS.get(pooling_name)(in_channels=pooling_in_channels, **pooling_params)

        # HEAD
        head_name = self._hparams.task.params.get('head_name')
        head_params = self._hparams.task.params.get('head_params', dict())
        head_in_channels = self.pooling.out_channels
        self.head = HEADS.get(head_name)(in_channels=head_in_channels, **head_params)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward method."""
        x = self.backbone(x)
        if self.neck is not None:
            x = self.neck(x)
        x = self.pooling(x)
        x = self.head(x)
        return x

    def forward_with_gt(self, batch: Dict[str, Union[torch.Tensor, int]]) -> Dict[str, torch.Tensor]:
        """Forward with ground truth labels."""
        input_data = batch['image']
        target = batch['target']
        features = self.backbone(input_data)
        if self.neck is not None:
            features = self.neck(features)
        embeddings = self.pooling(features)
        prediction = self.head(embeddings, target)
        output = {'target': target, 'embeddings': embeddings, 'prediction': prediction}
        return output



"""TorchOK HRNet.

Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/hrnet.py
Copyright 2019 Ross Wightman
Licensed under The Apache 2.0 License [see LICENSE for details]
"""
from typing import Any, Dict, List, Union

import torch
import torch.nn as nn
from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.models.helpers import build_model_with_cfg
from timm.models.hrnet import _BN_MOMENTUM, BasicBlock, blocks_dict, Bottleneck, cfg_cls, HighResolutionModule
from torch import Tensor

from torchok.constructor import BACKBONES
from torchok.models.backbones import BaseBackbone


def _cfg(url: str = '', **kwargs):
    return {
        'url': url,
        'input_size': (3, 224, 224),
        'pool_size': (7, 7),
        'crop_pct': 0.875,
        'interpolation': 'bilinear',
        'mean': IMAGENET_DEFAULT_MEAN,
        'std': IMAGENET_DEFAULT_STD,
        'first_conv': 'conv1',
        'classifier': 'classifier',
        **kwargs
    }


url_base_path = 'https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-hrnet/'
default_cfgs = {
    'hrnet_w18_small': _cfg(url=f'{url_base_path}/hrnet_w18_small_v1-f460c6bc.pth'),
    'hrnet_w18_small_v2': _cfg(url=f'{url_base_path}/hrnet_w18_small_v2-4c50a8cb.pth'),
    'hrnet_w18': _cfg(url=f'{url_base_path}/hrnetv2_w18-8cb57bb9.pth'),
    'hrnet_w30': _cfg(url=f'{url_base_path}/hrnetv2_w30-8d7f8dab.pth'),
    'hrnet_w32': _cfg(url=f'{url_base_path}/hrnetv2_w32-90d8c5fb.pth'),
    'hrnet_w40': _cfg(url=f'{url_base_path}/hrnetv2_w40-7cd397a4.pth'),
    'hrnet_w44': _cfg(url=f'{url_base_path}/hrnetv2_w44-c9ac8c18.pth'),
    'hrnet_w48': _cfg(url=f'{url_base_path}/hrnetv2_w48-abd2e6ab.pth'),
    'hrnet_w64': _cfg(url=f'{url_base_path}/hrnetv2_w64-b47cc881.pth'),
}


class HighResolutionNet(BaseBackbone):
    """HighResolutionNet model."""

    def __init__(self, cfg: Dict[str, Any], in_channels: int = 3):
        """Init HighResolutionNet.

        Args:
            cfg: Model config.
            in_channels: Input channels.
        """
        super().__init__(in_channels=in_channels,
                         out_channels=cfg['STAGE4']['NUM_CHANNELS'])
        self._out_encoder_channels = cfg['STAGE4']['NUM_CHANNELS']

        stem_width = cfg['STEM_WIDTH']
        self.conv1 = nn.Conv2d(in_channels, stem_width, kernel_size=3, stride=2, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(stem_width, momentum=_BN_MOMENTUM)
        self.act1 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(stem_width, 64, kernel_size=3, stride=2, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(64, momentum=_BN_MOMENTUM)
        self.act2 = nn.ReLU(inplace=True)

        self.stage1_cfg = cfg['STAGE1']
        num_channels = self.stage1_cfg['NUM_CHANNELS'][0]
        block = blocks_dict[self.stage1_cfg['BLOCK']]
        num_blocks = self.stage1_cfg['NUM_BLOCKS'][0]
        self.layer1 = self.__make_layer(block, 64, num_channels, num_blocks)
        stage1_out_channel = block.expansion * num_channels

        self.stage2_cfg = cfg['STAGE2']
        num_channels = self.stage2_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage2_cfg['BLOCK']]
        num_channels = [num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition1 = self.__make_transition_layer([stage1_out_channel], num_channels)
        self.stage2, pre_stage_channels = self.__make_stage(self.stage2_cfg, num_channels)

        self.stage3_cfg = cfg['STAGE3']
        num_channels = self.stage3_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage3_cfg['BLOCK']]
        num_channels = [num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition2 = self.__make_transition_layer(pre_stage_channels, num_channels)
        self.stage3, pre_stage_channels = self.__make_stage(self.stage3_cfg, num_channels)

        self.stage4_cfg = cfg['STAGE4']
        num_channels = self.stage4_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage4_cfg['BLOCK']]
        num_channels = [num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition3 = self.__make_transition_layer(pre_stage_channels, num_channels)
        self.stage4, out_channels = self.__make_stage(self.stage4_cfg, num_channels, multi_scale_output=True)

        self.init_weights()

    @torch.jit.ignore
    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def __make_transition_layer(self, num_channels_pre_layer: List[int],
                                num_channels_cur_layer: List[int]) -> nn.Module:
        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(nn.Sequential(
                        nn.Conv2d(num_channels_pre_layer[i], num_channels_cur_layer[i], 3, 1, 1, bias=False),
                        nn.BatchNorm2d(num_channels_cur_layer[i], momentum=_BN_MOMENTUM),
                        nn.ReLU(inplace=True)))
                else:
                    transition_layers.append(nn.Identity())
            else:
                conv3x3s = []
                for j in range(i + 1 - num_branches_pre):
                    inchannels = num_channels_pre_layer[-1]
                    outchannels = num_channels_cur_layer[i] if j == i - num_branches_pre else inchannels
                    conv3x3s.append(nn.Sequential(
                        nn.Conv2d(inchannels, outchannels, 3, 2, 1, bias=False),
                        nn.BatchNorm2d(outchannels, momentum=_BN_MOMENTUM),
                        nn.ReLU(inplace=True)))
                transition_layers.append(nn.Sequential(*conv3x3s))

        return nn.ModuleList(transition_layers)

    def __make_layer(self, block: Union[Bottleneck, BasicBlock], in_channels: int,
                     out_channels: int, num_blocks: int, stride: int = 1) -> nn.Module:
        """The method creates layer in the HRNet.

        Args:
            block: Block type.
            in_channels: Number of input channels.
            out_channels: Number of output channels.
            num_blocks: Number of blocks.
            stride: Stride.
        """
        downsample = None

        if stride != 1 or in_channels != out_channels * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(in_channels, out_channels * block.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(out_channels * block.expansion, momentum=_BN_MOMENTUM),
            )

        layers = [block(in_channels, out_channels, stride, downsample)]
        in_channels = out_channels * block.expansion
        for i in range(1, num_blocks):
            layers.append(block(in_channels, out_channels))

        return nn.Sequential(*layers)

    def __make_stage(self, layer_config: Dict[str, Any], in_channels: List[int],
                     multi_scale_output: bool = True) -> nn.Module:
        """The method creates stage in the HRNet.

        Args:
            layer_config: Layer configurations.
            in_channels: Number of input channels.
        """
        num_modules = layer_config['NUM_MODULES']
        num_branches = layer_config['NUM_BRANCHES']
        num_blocks = layer_config['NUM_BLOCKS']
        num_channels = layer_config['NUM_CHANNELS']
        block = blocks_dict[layer_config['BLOCK']]
        fuse_method = layer_config['FUSE_METHOD']

        modules = []
        for i in range(num_modules):
            # multi_scale_output is only used last module
            reset_multi_scale_output = multi_scale_output or i < num_modules - 1
            modules.append(HighResolutionModule(
                num_branches, block, num_blocks, in_channels, num_channels, fuse_method, reset_multi_scale_output)
            )
            in_channels = modules[-1].get_num_in_chs()

        return nn.Sequential(*modules), in_channels

    def forward_stages(self, x: Tensor) -> List[Tensor]:
        """The method forward the tensor through all stages.

        Args:
            x: Input tensor.
        """
        x = self.layer1(x)

        xl = [t(x) for i, t in enumerate(self.transition1)]
        yl = self.stage2(xl)

        xl = [t(yl[-1]) if not isinstance(t, nn.Identity) else yl[i] for i, t in enumerate(self.transition2)]
        yl = self.stage3(xl)

        xl = [t(yl[-1]) if not isinstance(t, nn.Identity) else yl[i] for i, t in enumerate(self.transition3)]
        yl = self.stage4(xl)
        return yl

    def forward(self, x: Tensor) -> List[Tensor]:
        """Forward method.

        Args:
            x: Input tensor.
        """
        x = self.forward_stem(x)
        yl = self.forward_stages(x)

        return yl

    def forward_stem(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.act1(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.act2(x)
        return x

    def forward_features(self, x: Tensor) -> List[Tensor]:
        """Forward backbone features and input tensor.

        Args:
            x: Input tensor.
        """
        return [x] + self.forward(x)


def create_hrnet(variant: str, pretrained: bool = False, **model_kwargs):
    """Create HighResolutionNet base model.

    Args:
        variant: Backbone type.
        pretrained: If True the pretrained weights will be loaded.
        model_kwargs: Kwargs for model (for example in_channels).
    """
    return build_model_with_cfg(HighResolutionNet, variant, pretrained, model_cfg=cfg_cls[variant],
                                pretrained_strict=False, kwargs_filter=('num_classes', 'global_pool', 'in_chans'),
                                **model_kwargs)


@BACKBONES.register_class
def hrnet_w18_small(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w18_small model."""
    return create_hrnet('hrnet_w18_small', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w18_small_v2(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w18_small_v2 model."""
    return create_hrnet('hrnet_w18_small_v2', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w18(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w18 model."""
    return create_hrnet('hrnet_w18', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w30(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w30 model."""
    return create_hrnet('hrnet_w30', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w32(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w32 model."""
    return create_hrnet('hrnet_w32', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w40(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w40 model."""
    return create_hrnet('hrnet_w40', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w44(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w44 model."""
    return create_hrnet('hrnet_w44', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w48(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w48 model."""
    return create_hrnet('hrnet_w48', pretrained, **kwargs)


@BACKBONES.register_class
def hrnet_w64(pretrained: bool = False, **kwargs):
    """It's constructing a hrnet_w64 model."""
    return create_hrnet('hrnet_w64', pretrained, **kwargs)



from pathlib import Path
from typing import Optional, Tuple, Union

import torch
import pandas as pd

from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose

from torchok.data.datasets.base import ImageDataset


class RetrievalDataset(ImageDataset):
    """Dataset for image retrieval validation.

    The searches are made by queries while looking for relevant items in the whole set of items.
    Where gallery items are treated non-relevant.

    Example matches csv:
    Query ids should be unique int values,
    otherwise the rows having the same query id will be treated as different matches.

    Relevant ids can be repeated in different queries.

    Scores reflect the order of similarity of the image to the query,
    a higher score corresponds to a greater similarity(must be float value > 0.).

    .. csv-table:: Match csv example
        :header: query, relevant, scores

        1194917,601566 554492 224125 2001716519,4 3 2 2
        1257924,456490,4

    Example img_list csv:
    img_list.csv maps the id's of query and relevant elements to image paths

    .. csv-table:: Image csv example
        :header: id, image_path

        1194917,data/img_1.jpg
        601566,data/img_2.jpg
        554492,data/img_3.jpg
        224125,data/img_4.jpg
        2001716519,data/img_5.jpg
        1257924,data/img_6.jpg
        456490,data/img_7.jpg

    .. csv-table:: Gallery Image csv example
        :header: id, image_paths

        8,data/db/img_1.jpg
        10,data/db/img_2.jpg
        12,data/db/img_3.jpg
    """

    def __init__(self,
                 data_folder: str,
                 matches_csv_path: str,
                 img_list_csv_path: str,
                 transform: Union[BasicTransform, BaseCompose],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 gallery_folder: Optional[str] = None,
                 gallery_list_csv_path: Optional[str] = None,
                 image_dtype: str = 'float32',
                 img_list_map_column: dict = None,
                 matches_map_column: dict = None,
                 gallery_map_column: dict = None,
                 grayscale: bool = False):
        """Init RetrievalDataset class.

        Args:
            data_folder: Directory with all the images.
            matches_csv_path: path to csv file where queries with their relevance scores are specified
            img_list_csv_path: path to mapping image identifiers to image paths. Format: id | path.
                Id from matches csv are linked to id from img_list csv
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations` library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations` library.
            gallery_folder: Path to a folder with all gallery images (traversed recursively).
                            When the gallery not specified all the remaining queries and relevant
                            will be considered as negative samples to a given query-relevant set.
            gallery_list_csv_path: Path to mapping image identifiers to image paths. Format: id | path.
            image_dtype: Data type of of the torch tensors related to the image.
            img_list_map_column: Image mapping column names. Key - TorchOk column name, Value - csv column name.
                default value: {'image_path': 'image_path', 'img_id': 'id'}
            matches_map_column: Matches mapping column names. Key - TorchOk column name, Value - csv column name.
                default value: {'query': 'query', 'relevant': 'relevant', 'scores': 'scores'}
            gallery_map_column: Gallery mapping column names. Key - TorchOk column name, Value - csv column name.
                default value: {'gallery_path': 'image_path', 'gallery_id': 'id'}
            grayscale: If True, image will be read as grayscale otherwise as RGB.

        Raises:
            ValueError: if gallery_folder True, but gallery_list_csv_path is None
        """
        super().__init__(transform, augment, image_dtype, grayscale)
        self.__data_folder = Path(data_folder)
        self.__matches_map_column = matches_map_column if matches_map_column is not None\
            else {'query': 'query',
                  'relevant': 'relevant',
                  'scores': 'scores'}

        self.__img_list_map_column = img_list_map_column if img_list_map_column is not None\
            else {'image_path': 'image_path',
                  'img_id': 'id'}

        self.__gallery_map_column = gallery_map_column if gallery_map_column is not None\
            else {'gallery_path': 'image_path',
                  'gallery_id': 'id'}

        self.__matches = pd.read_csv(self.__data_folder / matches_csv_path,
                                     usecols=[self.__matches_map_column['query'],
                                              self.__matches_map_column['relevant'],
                                              self.__matches_map_column['scores']],
                                     dtype={self.__matches_map_column['query']: int,
                                            self.__matches_map_column['relevant']: str,
                                            self.__matches_map_column['scores']: str})

        self.__img_paths = pd.read_csv(self.__data_folder / img_list_csv_path,
                                       usecols=[self.__img_list_map_column['img_id'],
                                                self.__img_list_map_column['image_path']],
                                       dtype={self.__img_list_map_column['img_id']: int,
                                              self.__img_list_map_column['image_path']: str})

        self.__n_relevant, self.__n_queries, self.__index2imgid,\
            self.__relevant_arr, self.__relevance_scores = self.__parse_match_csv()

        self._imgid2path = dict(zip(self.__img_paths[self.__img_list_map_column['img_id']],
                                    self.__img_paths[self.__img_list_map_column['image_path']]))

        if len(self._imgid2path) != len(self.__img_paths):
            raise ValueError('Image csv have the same id for different image paths')

        # filtering (save only img_id that in match.csv)
        self._imgid2path = {img_id: self._imgid2path[img_id] for img_id in self.__index2imgid.values()}
        self._data_len = self.__n_queries + self.__n_relevant

        if gallery_folder is not None:
            self.__gallery_folder = Path(gallery_folder)
            self.__gallery_list_csv_path = gallery_list_csv_path
            if self.__gallery_list_csv_path is None:
                raise ValueError('Argument `gallery_list_csv_path` is None, please send path to gallery_list_csv_path')

            self.__gallery_paths = pd.read_csv(self.__gallery_folder / self.__gallery_list_csv_path,
                                               usecols=[self.__gallery_map_column['gallery_id'],
                                                        self.__gallery_map_column['gallery_path']],
                                               dtype={self.__gallery_map_column['gallery_id']: int,
                                                      self.__gallery_map_column['gallery_path']: str})
            self.__gallery_imgid2path = dict(zip(self.__gallery_paths[self.__gallery_map_column['gallery_id']],
                                                 self.__gallery_paths[self.__gallery_map_column['gallery_path']]))

            if len(self.__gallery_imgid2path) != len(self.__gallery_paths):
                raise ValueError('Gallery csv have the same id for different image paths')

            self.__n_gallery = 0
            self._gallery_index2imgid = {}
            for img_id in self.__gallery_imgid2path:
                self._gallery_index2imgid[self._data_len + self.__n_gallery] = img_id
                self.__n_gallery += 1

            self._data_len += self.__n_gallery

        self.__imgid2index = {}
        for index, img_id in self.__index2imgid.items():
            if img_id not in self.__imgid2index:
                self.__imgid2index[img_id] = index

        self.__scores, self.__is_query = self.__get_targets()

    def __getitem__(self, index: int) -> dict:
        """Get item sample.

        Returns:
            sample: dict, where
            sample['image'] - Tensor, representing image after augmentations and transformations, dtype=image_dtype.
            sample['index'] - Index.
            sample['is_query'] - Int tensor, if item is query: return index of this query in target matrix, else -1.
            sample['scores'] - Float tensor shape (1, len(n_query)), relevant scores of current item.
        """
        if index < self.__n_queries + self.__n_relevant:
            img_id = self.__index2imgid[index]
            image_path = self.__data_folder / self._imgid2path[img_id]
        else:
            img_id = self._gallery_index2imgid[index]
            image_path = self.__gallery_folder / self.__gallery_imgid2path[img_id]

        image = self._read_image(image_path)
        sample = {'image': image}
        sample = self._apply_transform(self.augment, sample)
        sample = self._apply_transform(self.transform, sample)
        sample['image'] = sample['image'].type(torch.__dict__[self._image_dtype])
        sample['index'] = index
        sample['is_query'] = self.__is_query[index]
        sample['scores'] = self.__scores[index]

        return sample

    def __parse_match_csv(self) -> Tuple[int, int, dict, list, list]:
        query_arr = self.__matches.loc[:, 'query'].tolist()
        index2imgid = dict(enumerate(query_arr))
        n_queries = len(index2imgid)

        relevant_arr, relevance_scores = [], []
        n_relevant = 0

        for index in range(len(self.__matches)):
            row_relevants, row_scores = [], []
            rel_img_idxs = list(map(int, self.__matches.iloc[index]['relevant'].split()))
            rel_img_scores = list(map(float, self.__matches.iloc[index]['scores'].split()))

            if len(rel_img_idxs) != len(rel_img_scores):
                raise ValueError(f'Relevant objects list must match relevance scores list in size.'
                                 f'Got number of relevant object indices: {len(rel_img_idxs)}, '
                                 f'number of relevance scores: {len(rel_img_scores)}')

            for img_id, img_score in zip(rel_img_idxs, rel_img_scores):
                # save unique image_id
                if img_id not in index2imgid.values():
                    index2imgid[n_queries + n_relevant] = img_id
                    n_relevant += 1
                row_relevants.append(img_id)
                row_scores.append(img_score)

            relevant_arr.append(row_relevants)
            relevance_scores.append(row_scores)
        return n_relevant, n_queries, index2imgid, relevant_arr, relevance_scores

    def __get_targets(self) -> Tuple[torch.FloatTensor, torch.IntTensor]:
        """Maping item scores to queues.

        Returns:
            Two target tensor: scores and is_query.
            Scores is tensor with shape: (len(self), n_queries).
            Is_query is tensor with shape: (len(self)).

        Raises:
            ValueError: If relevant objects list doesn't match with relevance scores list in size.
        """
        scores = torch.zeros((len(self), self.__n_queries), dtype=torch.float32)
        is_query = torch.full((len(self),), -1, dtype=torch.int32)

        for index in range(self.__n_queries):
            relevant_img_idxs = self.__relevant_arr[index]
            relevance_scores = self.__relevance_scores[index]
            relevant_indices = [self.__imgid2index[img_id] for img_id in relevant_img_idxs]
            for rel_index, score in zip(relevant_indices, relevance_scores):
                scores[rel_index][index] = score
            is_query[index] = index

        return scores, is_query

    def __len__(self) -> int:
        """Length of Retrieval dataset."""
        return self._data_len

    @property
    def data_folder(self) -> str:
        """Directory with all the images."""
        return self.__data_folder

    @property
    def gallery_folder(self) -> Optional[str]:
        """Directory with all gallery images."""
        return self.__gallery_folder

    @property
    def n_queries(self) -> int:
        """Is number of queries."""
        return self.__n_queries

    @property
    def n_relevant(self) -> int:
        """Is number of relevant items."""
        return self.__n_relevant

    @property
    def n_gallery(self) -> Optional[int]:
        """Is number of gallery items."""
        return self.__n_gallery

    @property
    def scores(self) -> torch.FloatTensor:
        """Tensor of all scores."""
        return self.__scores

    @property
    def is_query(self) -> torch.IntTensor:
        """Tensor of all types of items."""
        return self.__is_query



from torchok.models.backbones.base_backbone import BaseBackbone
from torchok.models.backbones import efficientnet
from torchok.models.backbones import hrnet
from torchok.models.backbones import mobilenetv3
from torchok.models.backbones import resnet
from torchok.models.backbones import swin
from torchok.models.backbones import davit



"""
This HRNet implementation is modified from the following repository:
https://github.com/HRNet/HRNet-Semantic-Segmentation
"""
from functools import partial
from typing import Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from torchok.constructor import HEADS
from torchok.models.base import BaseModel
from torchok.models.modules.bricks.convbnact import ConvBnAct

resize = partial(F.interpolate, mode='bilinear', align_corners=True)
ConvBnRelu = partial(ConvBnAct, act_layer=nn.ReLU)
BN_MOMENTUM = 0.1


class SpatialGather_Module(nn.Module):
    """Aggregate the context features according to the initial predicted probability distribution.
    Employ the soft-weighted method to aggregate the context.
    """
    def __init__(self, num_classes: int = 0, scale: int = 1):
        """Init ObjectAttentionBlock.
        Args:
            num_classes: Number of classes.
            scale: Scale to downsample the input feature maps (save memory cost).
        """
        super(SpatialGather_Module, self).__init__()
        self.num_classes = num_classes
        self.scale = scale

    def forward(self, feats: Tensor, probs: Tensor) -> Tensor:
        batch_size, c, h, w = probs.size(0), probs.size(1), probs.size(2), probs.size(3)
        probs = probs.view(batch_size, c, -1)
        feats = feats.view(batch_size, feats.size(1), -1)
        feats = feats.permute(0, 2, 1)  # batch x hw x c
        probs = F.softmax(self.scale * probs, dim=2)  # batch x k x hw
        ocr_context = torch.matmul(probs, feats) \
            .permute(0, 2, 1).unsqueeze(3)  # batch x k x c
        return ocr_context


class ObjectAttentionBlock(nn.Module):
    """The basic implementation for object context block."""
    def __init__(self, in_channels: int, key_channels: int, scale: int = 1):
        """Init ObjectAttentionBlock.
        Args:
            in_channels: The dimension of the input feature map.
            key_channels: Number of channels in the dimension after the key/query transform.
            scale: Scale to downsample the input feature maps (save memory cost).
        """
        super(ObjectAttentionBlock, self).__init__()
        self.scale = scale
        self.in_channels = in_channels
        self.key_channels = key_channels
        self.pool = nn.MaxPool2d(kernel_size=(scale, scale))
        self.f_pixel = nn.Sequential(
            ConvBnRelu(self.in_channels, self.key_channels, kernel_size=1),
            ConvBnRelu(self.key_channels, self.key_channels, kernel_size=1)
        )
        self.f_object = nn.Sequential(
            ConvBnRelu(self.in_channels, self.key_channels, kernel_size=1),
            ConvBnRelu(self.key_channels, self.key_channels, kernel_size=1)
        )
        self.f_down = nn.Sequential(
            ConvBnRelu(self.in_channels, self.key_channels, kernel_size=1),
            ConvBnRelu(self.key_channels, self.key_channels, kernel_size=1)
        )
        self.f_up = ConvBnRelu(self.key_channels, self.in_channels, kernel_size=1)

    def forward(self, x: Tensor, proxy: Tensor) -> Tensor:
        batch_size, h, w = x.size(0), x.size(2), x.size(3)
        if self.scale > 1:
            x = self.pool(x)

        query = self.f_pixel(x).view(batch_size, self.key_channels, -1)
        query = query.permute(0, 2, 1)
        key = self.f_object(proxy).view(batch_size, self.key_channels, -1)
        value = self.f_down(proxy).view(batch_size, self.key_channels, -1)
        value = value.permute(0, 2, 1)

        sim_map = torch.matmul(query, key)
        sim_map = (self.key_channels ** -.5) * sim_map
        sim_map = F.softmax(sim_map, dim=-1)

        # add bg context ...
        context = torch.matmul(sim_map, value)
        context = context.permute(0, 2, 1).contiguous()
        context = context.view(batch_size, self.key_channels, *x.size()[2:])
        context = self.f_up(context)
        if self.scale > 1:
            context = resize(input=context, size=(h, w))

        return context


class SpatialOCR(nn.Module):
    """Implementation of the OCR module: Aggregate the global object representation to update the
    representation for each pixel.
    """

    def __init__(self, in_channels: int, key_channels: int, out_channels: int, scale: int = 1, dropout: float = 0.1):
        """Init SpatialOCR.
        Args:
            in_channels: Number of input channels.
            key_channels: Number of channels in the dimension after the key/query transform.
            out_channels: Number of output channels.
            scale: Scale to downsample the input feature maps (save memory cost).
            dropout: Dropout probability.
        """
        super(SpatialOCR, self).__init__()
        self.object_context_block = ObjectAttentionBlock(in_channels, key_channels, scale)
        _in_channels = 2 * in_channels

        self.conv_bn_dropout = nn.Sequential(
            ConvBnRelu(_in_channels, out_channels, kernel_size=1),
            nn.Dropout2d(dropout)
        )

    def forward(self, feats: Tensor, proxy_feats: Tensor) -> Tensor:
        context = self.object_context_block(feats, proxy_feats)
        output = self.conv_bn_dropout(torch.cat([context, feats], 1))

        return output


@HEADS.register_class
class HRNetSegmentationHead(BaseModel):
    """HRNet segmentation head."""

    def __init__(self, in_channels: int, num_classes: int):
        """Init HRNetSegmentationHead.
        Args:
            in_channels: Size of each input sample.
            num_classes: Number of classes.
        """
        super().__init__(in_channels, out_channels=num_classes)
        self.num_classes = num_classes
        self.convbnact = ConvBnRelu(in_channels,
                                    in_channels,
                                    kernel_size=1,
                                    padding=0,
                                    stride=1)
        self.final_conv_layer = nn.Conv2d(in_channels=in_channels,
                                          out_channels=num_classes,
                                          kernel_size=1,
                                          stride=1,
                                          padding=0)

    def forward(self, feats: Tensor) -> Tensor:
        input_image, feats = feats
        x = self.convbnact(feats)
        out = self.final_conv_layer(x)
        out = resize(out, input_image.shape[2:])
        return out


@HEADS.register_class
class HRNetOCRSegmentationHead(BaseModel):
    """
    Implementation of HRNet segmentation head with Object-Contextual Representations for Semantic Segmentation
    from https://github.com/HRNet/HRNet-Semantic-Segmentation/tree/HRNet-OCR.
    """

    has_ocr = True

    def __init__(self, in_channels: int, num_classes: int, ocr_mid_channels=128, ocr_key_channels=64):
        """Init HRNetOCRSegmentationHead.
        Implementation of HRNet with Object-Contextual Representations for Semantic Segmentation
        from https://github.com/HRNet/HRNet-Semantic-Segmentation/tree/HRNet-OCR.
        Args:
            in_channels: Number of channels from HRNetSegmentationNeck.
            num_classes: Number of segmentation classes.
            ocr_mid_channels: Number of intermediate feature channels.
            ocr_key_channels: Number of channels in the dimension after the key/query transform.
        """
        super(HRNetOCRSegmentationHead, self).__init__(in_channels, num_classes)
        self.num_classes = num_classes

        self.conv3x3_ocr = ConvBnRelu(in_channels, ocr_mid_channels, kernel_size=3, padding=1)
        self.ocr_gather_head = SpatialGather_Module(num_classes)

        self.ocr_distri_head = SpatialOCR(in_channels=ocr_mid_channels,
                                          key_channels=ocr_key_channels,
                                          out_channels=ocr_mid_channels,
                                          scale=1,
                                          dropout=0.05,
                                          )
        self.last_reduction = ConvBnRelu(
            ocr_mid_channels,
            ocr_mid_channels // 16,
            kernel_size=1, stride=1, padding=0
        )

        self.cls_head = nn.Conv2d(ocr_mid_channels // 16, num_classes,
                                  kernel_size=1, stride=1, padding=0, bias=True)

        self.aux_head = nn.Sequential(
            ConvBnRelu(in_channels, in_channels,
                       kernel_size=1, stride=1, padding=0),
            nn.Conv2d(in_channels, num_classes,
                      kernel_size=1, stride=1, padding=0, bias=True)
        )

        self.init_weights()

    def init_weights(self):
        for name, m in self.named_modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight, mode="fan_in", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, feats: Tensor) -> Union[Tensor, Tuple[Tensor, Tensor]]:
        input_image, feats = feats

        out_aux = self.aux_head(feats)
        # compute contrast feature
        feats = self.conv3x3_ocr(feats)

        context = self.ocr_gather_head(feats, out_aux)
        feats = self.ocr_distri_head(feats, context)

        feats = self.last_reduction(feats)
        out = self.cls_head(feats)

        out = resize(out, input_image.shape[2:])
        out_aux = resize(out_aux, input_image.shape[2:])

        if self.num_classes == 1:
            out = out[:, 0]
            out_aux = out_aux[:, 0]

        if self.training:
            return out, out_aux
        else:
            return out



"""TorchOK DaViT.

Adapted from https://github.com/dingmyu/davit/blob/main/mmseg/mmseg/models/backbones/davit.py
Licensed under MIT License [see LICENSE for details]
"""
import itertools
from typing import Optional, Tuple, List

import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.models.helpers import build_model_with_cfg
from timm.models.layers import DropPath, trunc_normal_
from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from torch import Tensor

from torchok.constructor import BACKBONES
from torchok.models.base import BaseModel
from torchok.models.modules.bricks.mlp import Mlp


def _cfg(url='', **kwargs):
    return {
        'url': url, 'input_size': (3, 224, 224), 'pool_size': None,
        'crop_pct': .9, 'interpolation': 'bicubic', 'fixed_input_size': True,
        'mean': IMAGENET_DEFAULT_MEAN, 'std': IMAGENET_DEFAULT_STD,
        # 'first_conv': 'patch_embed.proj', 'classifier': 'head',
        **kwargs
    }


default_cfgs = {
    'davit_t': _cfg(url='https://torchok-hub.s3.eu-west-1.amazonaws.com/davit-t_torchok.pth'),
    'davit_s': _cfg(url='https://torchok-hub.s3.eu-west-1.amazonaws.com/davit-s_torchok.pth'),
    'davit_b': _cfg(url='https://torchok-hub.s3.eu-west-1.amazonaws.com/davit-b_torchok.pth'),
}


class PatchEmbed(nn.Module):
    """2D Image to Patch Embedding."""

    def __init__(self, patch_size: int = 16, in_channels: int = 3, embed_dim: int = 96, overlapped: bool = False):
        """Init PatchEmbed.

        Args:
            patch_size: Patch size.
            in_channels: Input channels.
            embed_dim: Embedding dimension.
            overlapped: Overlapping.
        """
        super().__init__()
        self.patch_size = patch_size

        if patch_size == 4:
            self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=(7, 7), stride=patch_size, padding=(3, 3))
            self.norm = nn.LayerNorm(embed_dim)
        elif patch_size == 2:
            kernel = 3 if overlapped else 2
            pad = 1 if overlapped else 0
            self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=kernel, stride=patch_size, padding=pad)
            self.norm = nn.LayerNorm(in_channels)

    def forward(self, x: Tensor, size: Tuple[int]) -> Tuple[Tensor, Tuple[int]]:
        """Forward method."""
        H, W = size
        dim = len(x.shape)
        if dim == 3:
            B, HW, C = x.shape
            x = self.norm(x)
            x = x.reshape(B, H, W, C).permute(0, 3, 1, 2).contiguous()

        B, C, H, W = x.shape
        if W % self.patch_size != 0:
            x = F.pad(x, (0, self.patch_size - W % self.patch_size[1]))
        if H % self.patch_size != 0:
            x = F.pad(x, (0, 0, 0, self.patch_size - H % self.patch_size[0]))

        x = self.proj(x)
        newsize = (x.size(2), x.size(3))
        x = x.flatten(2).transpose(1, 2)
        if dim == 4:
            x = self.norm(x)
        return x, newsize


class ConvPosEnc(nn.Module):
    """Convolution positional encoding."""

    def __init__(self, dim: int, kernel_size: int = 3, use_act: bool = False, norm_layer: Optional[str] = None):
        """Init ConvPosEnc.

        Args:
            dim: Dimension.
            kernel_size: Kernel size.
            use_act: If True, will use GELU activation.
            norm_layer: Type of normalization.
        """
        super(ConvPosEnc, self).__init__()
        self.proj = nn.Conv2d(dim,
                              dim,
                              kernel_size,
                              1,
                              kernel_size // 2,
                              groups=dim)
        self.norm_layer = norm_layer
        if self.norm_layer == 'batch':
            self.norm = nn.BatchNorm2d(dim)
        elif self.norm_layer == 'layer':
            self.norm = nn.LayerNorm(dim)
        self.activation = nn.GELU() if use_act else None

    def forward(self, x: Tensor, size: Tuple[int, int]) -> Tensor:
        """Forward method."""
        B, N, C = x.shape
        H, W = size

        feat = x.transpose(1, 2).view(B, C, H, W)
        feat = self.proj(feat)
        if self.norm_layer == 'batch':
            feat = self.norm(feat).flatten(2).transpose(1, 2)
        elif self.norm_layer == 'layer':
            feat = self.norm(feat.flatten(2).transpose(1, 2))
        else:
            feat = feat.flatten(2).transpose(1, 2)

        if self.activation is not None:
            x = x + self.activation(feat)
        return x


class ChannelAttention(nn.Module):
    """Channel based multi-head self attention module."""

    def __init__(self, dim: int, num_heads: int = 8, qkv_bias: bool = False):
        """Init Channel Attention.

        Args:
            dim: Dimension size.
            num_heads: Number of heads.
            qkv_bias: Query-Key-Value bias.
        """
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x: Tensor) -> Tensor:
        """Forward method."""
        B, N, C = x.shape

        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        k = k * self.scale
        attention = k.transpose(-1, -2) @ v
        attention = attention.softmax(dim=-1)
        x = (attention @ q.transpose(-1, -2)).transpose(-1, -2)
        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        return x


class WindowAttention(nn.Module):
    """Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window."""

    def __init__(self, dim: int, num_heads: int, qkv_bias: bool = True):
        """Init WindowAttention.

        Args:
            dim: Number of input channels.
            num_heads: Number of attention heads.
            qkv_bias:  If True, add a learnable bias to query, key, value. Default: True
        """

        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x: Tensor) -> Tensor:
        """Forward method.

        Args:
            x: Input tensor.
        """
        B_, N, C = x.shape
        qkv = self.qkv(x).reshape(B_, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))
        attn = self.softmax(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.proj(x)
        return x


class ChannelBlock(nn.Module):
    """Channel Block of DaViT."""

    def __init__(self, dim: int, num_heads: int, mlp_ratio: float = 4., qkv_bias: bool = False,
                 drop_path: float = 0., act_layer: nn.Module = nn.GELU, norm_layer: nn.Module = nn.LayerNorm,
                 ffn: bool = True, cpe_act: bool = False):
        """Init ChannelBlock.

        Args:
            dim: Dimension.
            num_heads: Number of heads.
            mlp_ratio: Multilayer perceptron ratio for hidden dim.
            qkv_bias: Query-Key-Value bias.
            drop_path: Drop path.
            act_layer: Activation layer.
            norm_layer: Normalization layer.
            ffn: If True, will use Mlp.
            cpe_act: If True, ConvPosEnc will use activation.

        """
        super().__init__()

        self.cpe = nn.ModuleList([ConvPosEnc(dim=dim, kernel_size=3, use_act=cpe_act),
                                  ConvPosEnc(dim=dim, kernel_size=3, use_act=cpe_act)])
        self.ffn = ffn
        self.norm1 = norm_layer(dim)
        self.attn = ChannelAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        if self.ffn:
            self.norm2 = norm_layer(dim)
            mlp_hidden_dim = int(dim * mlp_ratio)
            self.mlp = Mlp(
                in_features=dim,
                hidden_features=mlp_hidden_dim,
                act_layer=act_layer)

    def forward(self, inputs):
        """Forward method."""
        x, size = inputs
        x = self.cpe[0](x, size)
        cur = self.norm1(x)
        cur = self.attn(cur)
        x = x + self.drop_path(cur)

        x = self.cpe[1](x, size)
        if self.ffn:
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x, size


class SpatialBlock(nn.Module):
    """Windows Block.

    Args:
        dim: Number of input channels.
        num_heads: Number of attention heads.
        window_size: Window size.
        mlp_ratio: Ratio of mlp hidden dim to embedding dim.
        qkv_bias: If True, add a learnable bias to query, key, value. Default: True
        drop_path: Stochastic depth rate. Default: 0.0
        act_layer: Activation layer. Default: nn.GELU
        norm_layer: Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, dim: int, num_heads: int, window_size: int = 7, mlp_ratio: float = 4.,
                 qkv_bias: bool = True, drop_path: float = 0., act_layer: nn.Module = nn.GELU,
                 norm_layer: nn.Module = nn.LayerNorm, ffn: bool = True, cpe_act: bool = False):
        super().__init__()
        self.dim = dim
        self.ffn = ffn
        self.num_heads = num_heads
        self.window_size = window_size
        self.mlp_ratio = mlp_ratio
        self.cpe = nn.ModuleList([ConvPosEnc(dim=dim, kernel_size=3, use_act=cpe_act),
                                  ConvPosEnc(dim=dim, kernel_size=3, use_act=cpe_act)])

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        if self.ffn:
            self.norm2 = norm_layer(dim)
            mlp_hidden_dim = int(dim * mlp_ratio)
            self.mlp = Mlp(
                in_features=dim,
                hidden_features=mlp_hidden_dim,
                act_layer=act_layer)

    def forward(self, inputs):
        """Forward method."""
        x, size = inputs
        H, W = size
        B, L, C = x.shape

        shortcut = self.cpe[0](x, size)
        x = self.norm1(shortcut)
        x = x.view(B, H, W, C)

        pad_l = pad_t = 0
        pad_r = (self.window_size - W % self.window_size) % self.window_size
        pad_b = (self.window_size - H % self.window_size) % self.window_size
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape

        x_windows = self._window_partition(x, self.window_size)
        x_windows = x_windows.view(-1, self.window_size * self.window_size, C)

        # W-MSA/SW-MSA
        attn_windows = self.attn(x_windows)

        # merge windows
        attn_windows = attn_windows.view(-1,
                                         self.window_size,
                                         self.window_size,
                                         C)

        x = self._window_reverse(attn_windows, self.window_size, Hp, Wp)

        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()

        x = x.view(B, H * W, C)
        x = shortcut + self.drop_path(x)

        x = self.cpe[1](x, size)
        if self.ffn:
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x, size

    def _window_partition(self, x: Tensor, window_size: int) -> Tensor:
        """
        Args:
            x: (B, H, W, C)
            window_size (int): window size

        Returns:
            windows: (num_windows*B, window_size, window_size, C)
        """
        B, H, W, C = x.shape
        x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
        windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
        return windows

    def _window_reverse(self, windows: Tensor, window_size: int, H: int, W: int) -> Tensor:
        """
        Args:
            windows: (num_windows*B, window_size, window_size, C)
            window_size (int): Window size
            H (int): Height of image
            W (int): Width of image

        Returns:
            x: (B, H, W, C)
        """
        B = int(windows.shape[0] / (H * W / window_size / window_size))
        x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
        x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
        return x


class DaViT(BaseModel):
    """ Dual Attention Transformer"""

    def __init__(self, img_size: int = 224, in_channels: int = 3, patch_size: int = 4, depths=(1, 1, 3, 1),
                 embed_dims: Tuple[int] = (64, 128, 192, 256), num_heads: Tuple[int] = (3, 6, 12, 24),
                 window_size: int = 7, mlp_ratio: float = 4., qkv_bias: bool = True, drop_path_rate: float = 0.1,
                 norm_layer: nn.Module = nn.LayerNorm, overlapped_patch: bool = False, ffn: bool = True,
                 cpe_act: bool = False):
        """Init DaViT.

        Args:
            img_size: Input image size.
            in_channels: Number of input image channels.
            patch_size: Patch size.
            embed_dims: Patch embedding dimension.
            num_heads: Number of attention heads in different layers.
            window_size: Window size.
            mlp_ratio: Ratio of mlp hidden dim to embedding dim.
            qkv_bias: If True, add a learnable bias to query, key, value.
            drop_path_rate: Stochastic depth rate.
            norm_layer: Normalization layer.
        """
        super().__init__(in_channels, embed_dims[-1])
        self.img_size = img_size

        architecture = [[index] * item for index, item in enumerate(depths)]
        self.attention_types = ('spatial', 'channel')
        self.architecture = architecture
        self.embed_dims = embed_dims
        self.num_heads = num_heads
        self.num_stages = len(self.embed_dims)
        self._out_encoder_channels = embed_dims
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, 2 * len(list(itertools.chain(*self.architecture))))]

        self.patch_embeds = nn.ModuleList([
            PatchEmbed(patch_size=patch_size if i == 0 else 2,
                       in_channels=in_channels if i == 0 else self.embed_dims[i - 1],
                       embed_dim=self.embed_dims[i],
                       overlapped=overlapped_patch)
            for i in range(self.num_stages)])

        main_blocks = []
        for block_id, block_param in enumerate(self.architecture):
            layer_offset_id = len(list(itertools.chain(*self.architecture[:block_id])))

            block = nn.ModuleList([
                nn.Sequential(*[
                    ChannelBlock(
                        dim=self.embed_dims[item],
                        num_heads=self.num_heads[item],
                        mlp_ratio=mlp_ratio,
                        qkv_bias=qkv_bias,
                        drop_path=dpr[2 * (layer_id + layer_offset_id) + attention_id],
                        norm_layer=norm_layer,
                        ffn=ffn,
                        cpe_act=cpe_act
                    ) if attention_type == 'channel' else
                    SpatialBlock(
                        dim=self.embed_dims[item],
                        num_heads=self.num_heads[item],
                        mlp_ratio=mlp_ratio,
                        qkv_bias=qkv_bias,
                        drop_path=dpr[2 * (layer_id + layer_offset_id) + attention_id],
                        norm_layer=norm_layer,
                        ffn=ffn,
                        cpe_act=cpe_act,
                        window_size=window_size,
                    ) if attention_type == 'spatial' else None
                    for attention_id, attention_type in enumerate(self.attention_types)]
                              ) for layer_id, item in enumerate(block_param)
            ])
            main_blocks.append(block)
        self.main_blocks = nn.ModuleList(main_blocks)

        # add a norm layer for each output
        for i_layer in range(self.num_stages):
            layer = norm_layer(self.embed_dims[i_layer])  # if i_layer != 0 else nn.Identity()
            layer_name = f'norm{i_layer}'
            self.add_module(layer_name, layer)

        self.__init_weights()

    def __init_weights(self):
        """Weight initialization."""
        for m in self.modules():
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

    def _forward_stages(self, x: Tensor) -> Tuple[List[Tensor], List[Tensor]]:
        x, size = self.patch_embeds[0](x, (x.size(2), x.size(3)))
        features = [x]
        sizes = [size]
        branches = [0]

        for block_index, block_param in enumerate(self.architecture):
            branch_ids = sorted(set(block_param))
            for branch_id in branch_ids:
                if branch_id not in branches:
                    x, size = self.patch_embeds[branch_id](features[-1], sizes[-1])
                    features.append(x)
                    sizes.append(size)
                    branches.append(branch_id)
            for layer_index, branch_id in enumerate(block_param):
                inputs = (features[branch_id], sizes[branch_id])
                features[branch_id], _ = self.main_blocks[block_index][layer_index](inputs)
        return features, sizes

    def forward_features(self, x: Tensor) -> List[Tensor]:
        """Forward method for getting backbone feature maps.
           They are mainly used for segmentation and detection tasks.

        Args:
            x: Input tensor.

        Returns:
            This method return tuple of tensors.
            Each tensor has (B, C, H, W) shape structure.
        """
        input_tensor = x
        features, sizes = self._forward_stages(x)

        outs = []
        for i in range(self.num_stages):
            norm_layer = getattr(self, f'norm{i}')
            x_out = norm_layer(features[i])
            H, W = sizes[i]
            out = x_out.view(-1, H, W, self.embed_dims[i]).permute(0, 3, 1, 2).contiguous()
            outs.append(out)

        return [input_tensor, *outs]

    def forward(self, x: Tensor) -> Tensor:
        """Forward method"""
        features, sizes = self._forward_stages(x)
        last_stage = self.num_stages - 1

        norm_layer = getattr(self, f'norm{last_stage}')
        x_out = norm_layer(features[last_stage])
        H, W = sizes[last_stage]
        out = x_out.view(-1, H, W, self.embed_dims[last_stage]).permute(0, 3, 1, 2).contiguous()

        return out


def create_davit(variant: str, pretrained: bool = False, **kwargs):
    """Create DaViT base model.

    Args:
        variant: Backbone type.
        pretrained: If True the pretrained weights will be loaded.
        kwargs: Kwargs for model (for example in_chans).
    """
    kwargs_filter = ('num_classes', 'global_pool', 'in_chans')
    return build_model_with_cfg(DaViT, variant, pretrained, pretrained_strict=False,
                                pretrained_cfg=default_cfgs[variant], kwargs_filter=kwargs_filter, **kwargs)


@BACKBONES.register_class
def davit_t(pretrained: bool = False, **kwargs):
    """It's constructing a davit_t model."""
    model_kwargs = dict(embed_dims=(96, 192, 384, 768), depths=(1, 1, 3, 1), num_heads=(3, 6, 12, 24), **kwargs)
    return create_davit('davit_t', pretrained, **model_kwargs)


@BACKBONES.register_class
def davit_s(pretrained: bool = False, **kwargs):
    """It's constructing a davit_s model."""
    model_kwargs = dict(embed_dims=(96, 192, 384, 768), depths=(1, 1, 9, 1), num_heads=(3, 6, 12, 24), **kwargs)
    return create_davit('davit_s', pretrained, **model_kwargs)


@BACKBONES.register_class
def davit_b(pretrained: bool = False, **kwargs):
    """It's constructing a davit_b model."""
    model_kwargs = dict(embed_dims=(128, 256, 512, 1024), depths=(1, 1, 9, 1), num_heads=(4, 8, 16, 32), **kwargs)
    return create_davit('davit_b', pretrained, **model_kwargs)



from pathlib import Path
from typing import Union, Optional

import torch
import pandas as pd
from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose

from torchok.data.datasets.base import ImageDataset


class UnsupervisedContrastiveDataset(ImageDataset):
    """A dataset for unsupervised contrastive task.

    One image is transformed twice so that they are positive to each other.

    .. csv-table:: UnsupervisedContrastive csv example
        :header: image_path
        
        cat_1.jpg
        dog_1.jpg
    """

    def __init__(self,
                 data_folder: str,
                 csv_path: str,
                 transform: Union[BasicTransform, BaseCompose],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 image_dtype: str = 'float32',
                 csv_columns_mapping: dict = None,
                 grayscale: bool = False):
        """Init UnsupervisedContrastiveDataset.

        Args:
            data_folder: Directory with all the images.
            csv_path: Path to the csv file with path to images and annotations.
                Path to images must be under column `input_column`.
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations` library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations` library.
            image_dtype: data type of of the torch tensors related to the image.
            csv_columns_mapping: Matches mapping column names. Key - TorchOK column name, Value - csv column name.
                default value: {'image_path': 'image_path'}
            grayscale: if True image will be read as grayscale otherwise as RGB.
        """
        super().__init__(transform, augment, image_dtype, grayscale)
        self.__data_folder = Path(data_folder)
        self.__csv_path = csv_path
        self.__csv_columns_mapping = csv_columns_mapping if csv_columns_mapping is not None\
            else {'image_path': 'image_path'}
        self.__input_column = self.__csv_columns_mapping['image_path']
        self.__csv = pd.read_csv(self.__data_folder / self.__csv_path, dtype={self.__input_column: 'str'})

    def __getitem__(self, idx: int) -> dict:
        """Get item sample.

        Returns:
            sample: dict, where
            sample['image_0'] - Tensor, representing image after augmentations and transformations, dtype=image_dtype.
            sample['image_1'] - Tensor, representing image after augmentations and transformations, dtype=image_dtype.
            sample['index'] - Index.
        """
        record = self.__csv.iloc[idx]
        image_path = self.__data_folder / record[self.__input_column]
        image = self._read_image(image_path)
        sample = {'image': image}

        sample_0_transformed = self._apply_transform(self.augment, sample)['image']
        sample_1_transformed = self._apply_transform(self.augment, sample)['image']

        sample_0_augmented = self._apply_transform(self.transform, {'image': sample_0_transformed})
        sample_1_augmented = self._apply_transform(self.transform, {'image': sample_1_transformed})

        sample_0 = sample_0_augmented['image'].type(torch.__dict__[self._image_dtype])
        sample_1 = sample_1_augmented['image'].type(torch.__dict__[self._image_dtype])

        return {'image_0': sample_0, 'image_1': sample_1, 'index': idx}

    def __len__(self) -> int:
        """Dataset length."""
        return len(self.__csv)

    @property
    def csv_columns_mapping(self) -> dict:
        """Column name matching."""
        return self.__csv_columns_mapping



import torch
import torch.nn.functional as F

from torchok.constructor import POOLINGS
from torchok.models.base import BaseModel


def adaptive_avgmax_pool2d(x, output_size=1):
    x_avg = F.adaptive_avg_pool2d(x, output_size)
    x_max = F.adaptive_max_pool2d(x, output_size)
    return 0.5 * (x_avg + x_max)


def adaptive_catavgmax_pool2d(x, output_size=1):
    x_avg = F.adaptive_avg_pool2d(x, output_size)
    x_max = F.adaptive_max_pool2d(x, output_size)
    return torch.cat((x_avg, x_max), 1)


def select_adaptive_pool2d(x, pool_type='avg', flatten: bool = True, output_size=1):
    """Selectable global pooling function with dynamic input kernel size
    """
    if pool_type == 'avg':
        x = F.adaptive_avg_pool2d(x, output_size)
    elif pool_type == 'avgmax':
        x = adaptive_avgmax_pool2d(x, output_size)
    elif pool_type == 'catavgmax':
        x = adaptive_catavgmax_pool2d(x, output_size)
    elif pool_type == 'max':
        x = F.adaptive_max_pool2d(x, output_size)
    else:
        assert False, 'Invalid pool type: %s' % pool_type

    if flatten:
        x = x.flatten(1)

    return x


@POOLINGS.register_class
class Pooling(BaseModel):
    def __init__(self, in_channels: int, pooling_type: str = 'avg', output_size=1):
        self.output_size = output_size
        self.pooling_type = pooling_type
        out_channels = in_channels if pooling_type != 'catavgmax' else 2 * in_channels
        super().__init__(in_channels, out_channels)

    def forward(self, x):
        return select_adaptive_pool2d(x, self.pooling_type, self.output_size)



import torch
import torch.nn as nn

from torchok.constructor import POOLINGS
from torchok.models.base import BaseModel
from torchok.models.poolings.classification import Pooling


@POOLINGS.register_class
class PoolingLinear(BaseModel):
    def __init__(self, in_channels, out_channels, pooling_type='avg', bias=True):
        super().__init__(in_channels, out_channels)

        self.global_pool = Pooling(in_channels=in_channels, pooling_type=pooling_type)
        self.fc = nn.Linear(self.global_pool._out_channels, out_channels, bias=bias)
        self.init_weights()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.global_pool(x)
        x = self.fc(x)
        return x

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)



from abc import ABC

from sklearn.preprocessing import normalize
from typing import Callable, List, Optional, Union, Tuple, Generator

import torch
import faiss
import numpy as np
import pandas as pd
import math

from torchmetrics import Metric
from ranx.metrics import precision, recall, average_precision, ndcg, hit_rate
from enum import Enum

from torchok.constructor import METRICS


__all__ = [
    'PrecisionAtKMeter',
    'RecallAtKMeter',
    'MeanAveragePrecisionAtKMeter',
    'NDCGAtKMeter',
]


class DatasetType(Enum):
    CLASSIFICATION = 'classification'
    REPRESENTATION = 'representation'


class MetricDistance(Enum):
    IP = 'IP'
    L2 = 'L2'


dataset_enum_mapping = {
    'classification': DatasetType.CLASSIFICATION,
    'representation': DatasetType.REPRESENTATION,
}

distance_enum_mapping = {
    'IP': MetricDistance.IP,
    'L2': MetricDistance.L2
}


class IndexBasedMeter(Metric, ABC):
    """Base class for representation metrics.

    Store retrieval vectors and targets during phase in update method. FAISS library is used to build an index
    and search for top-k in it. Supports 2 datasets: classification dataset with targets,
    and representation dataset with scores and queries_idxs tensors.
    Compute method return generator with relevant and closest (FAISS searched) indexes. The relevant index
    contain it's relevant index with scores for current query. And the closest contain closest index with it's distance.
    """
    def __init__(self, exact_index: bool, dataset_type: str, metric_distance: str,
                 metric_func: Callable, k: Optional[int] = None, search_batch_size: Optional[int] = None,
                 normalize_vectors: bool = False, **kwargs):
        """Initialize IndexBasedMeter.

        Args:
            exact_index: If true then build fair inner product or Euclidean index (depends on the metric chosen),
                otherwise the index will be an approximate nearest neighbors search index.
            dataset_type: Dataset type (classification or representation), which will be used to calculate metric.
            metric_distance: Metric distance (IP - cosine distance, L2 - euclidean), which will be used to build
                FAISS index.
            metric_func: Representation metric (e.g ranx metric function) with the follow backend
                `def metric_func(
                    qrels: Union[np.ndarray, numba.typed.List],
                    run: Union[np.ndarray, numba.typed.List],
                    k: int = 0,
                ) -> np.ndarray`
                where qrels - y_true and run - y_pred
                see https://github.com/AmenRa/ranx/blob/ccab1549de81e7366e34213c86089e965db55f72/ranx/metrics.py
                for more details.
            k: Number of top closest indexes to get.
            search_batch_size: The size for one FAISS search request.
            normalize_vectors: If true vectors will be normalize, overwise no.

        Raises:
            ValueError: If metric or dataset is not correct write.
        """
        super().__init__(**kwargs)
        self.exact_index = exact_index

        self.dataset_type = dataset_enum_mapping[dataset_type]
        self.metric_distance = distance_enum_mapping[metric_distance]
        self.metric_func = metric_func
        self.normalize_vectors = normalize_vectors
        # set search_batch_size as num CPUs if search_batch_size is None
        self.search_batch_size = torch.get_num_threads() if search_batch_size is None else search_batch_size

        # set k as 1 if k is None
        k = 1 if k is None else k
        # as queries vectors can be relevant for another query vectors, so we should search k + 1 and delete first
        # and remove the first found vector because it would be query
        self.search_k = k + 1
        # but in metric compute study, k must not change
        self.metric_compute_k = k

        self.add_state('vectors', default=[], dist_reduce_fx=None)
        if self.dataset_type == DatasetType.CLASSIFICATION:
            # if classification dataset
            self.add_state('targets', default=[], dist_reduce_fx=None)
        else:
            # if representation dataset
            self.add_state('query_idxs', default=[], dist_reduce_fx=None)
            self.add_state('scores', default=[], dist_reduce_fx=None)

    def update(self, vectors: torch.Tensor, targets: Optional[torch.Tensor] = None,
               query_idxs: Optional[torch.Tensor] = None, scores: Optional[torch.Tensor] = None):
        """Append tensors in storage.

        Args:
            vectors: Often it would be embeddings, size (batch_size, embedding_size).
            targets: The labels for every vectors in classification mode, size (batch_size).
            query_idxs: Integer tensor where values >= 0 represent indices of queries with corresponding
                vectors in vectors tensor and value -1 indicates that the corresponding vector isn't a query.
            scores: The scores tensor, see representation dataset for more information,
                size (batch_size, total_num_queries).

        Raises:
            ValueError: If dataset is of classification type and targets is None, or if dataset is of representation
                type and at least one of scores or query_idxs is None.
        """
        vectors = vectors.detach().cpu()
        self.vectors.append(vectors)
        if self.dataset_type == DatasetType.CLASSIFICATION:
            if targets is None:
                raise ValueError("In classification dataset target must be not None.")
            targets = targets.detach().cpu()
            self.targets.append(targets)
        else:
            if query_idxs is None:
                raise ValueError("In representation dataset query_numbers must be not None.")
            if scores is None:
                raise ValueError("In representation dataset scores must be not None")

            query_idxs = query_idxs.detach().cpu()
            self.query_idxs.append(query_idxs)

            scores = scores.detach().cpu()
            self.scores.append(scores)

    def compute(self) -> float:
        """Compute metric value.

        Firstly it gathers all tensors in storage (done by torchmetrics).
        Then it prepares data, separates query and database vectors.
        Then it builds the FAISS index.
        Then, it create a generator of relevant and closest arrays.
        Finally, it compute metric.

        Returns:
            metric: Metric value.
        """
        vectors = torch.cat(self.vectors).numpy()
        if self.normalize_vectors:
            vectors = normalize(vectors)

        if self.dataset_type == DatasetType.CLASSIFICATION:
            # if classification dataset
            targets = torch.cat(self.targets).numpy()
            # prepare data
            relevant_idxs, gallery_idxs, query_row_idxs, query_as_relevant = self.prepare_classification_data(targets)
            # mock scores and query column indexes because it belong to representation data
            scores = None
            query_column_idxs = None
        else:
            # if representation dataset
            scores = torch.cat(self.scores).numpy()
            query_idxs = torch.cat(self.query_idxs).numpy()
            # prepare data
            relevant_idxs, gallery_idxs, query_column_idxs, \
                query_row_idxs, query_as_relevant = self.prepare_representation_data(query_idxs, scores)

        # build index
        vectors = vectors.astype(np.float32)
        index = self.build_index(vectors[gallery_idxs])

        # create relevant, closest generator
        generator = self.query_generator(index, vectors, relevant_idxs, query_row_idxs,
                                         gallery_idxs, query_as_relevant, self.search_k, scores, query_column_idxs)

        # compute metric
        metrics = []
        for relevant_idx, closest_idx in generator:
            metrics += self.metric_func(relevant_idx, closest_idx, k=self.metric_compute_k).tolist()
        metric = np.mean(metrics)
        return metric

    def prepare_representation_data(self,
                                    query_idxs: np.ndarray,
                                    scores: np.ndarray
                                    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Prepare data for faiss build index, and following search, in representation dataset case.

        Separate query and database vectors from storage vectors.
        Prepare scores.
        Generate relevant indexes for every query request.

        Args:
            query_idxs: Integer array where values >= 0 represent indices of queries with corresponding
                vectors in vectors tensor and value -1 indicates that the corresponding vector isn't a query. Also
                value of this array is the column number in the scores matrix, which need to reproduce
                relevant elements.
            scores: The scores tensor, see representation dataset for more information,
                size (batch_size, total_num_queries).
                Example of score matrix:
                    SCORES = torch.tensor(
                        [
                            [0, 0, 0],
                            [1, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 2, 0],
                            [0, 0, 1],
                            [0, 0, 2],
                            [0, 0, 4],
                            [0, 4, 0],
                        ]
                    ),
                it has 3 queries in [0, 2, 3] row position where only zeros, in order to find relevant vector index you
                need to look at the column, first column for first query, second column for second and so on. For query
                0 relevant vector index would be 1 with score 1, for query 1 - relevant = [4, 8] with scores = [2, 4]
                respectively, for query 2 - relevant = [5, 6, 7] with scores = [1, 2 , 4] respectively.

        Returns:
            relevant_idxs: Array of relevant indexes in gallery data, for every query.
            gallery_idxs: Array of gallery indexes in vectors storage.
            query_column_idxs: Array of queries column indexes in scores matrix. It means query order number in origin
                dataset, because after shuffle dataset the query order may be change we need to store it to reproduce
                for every query vector it's relevant.
            query_row_idxs: Array of queries row indexes in scores matrix, lso this index belong to vectors storage,
                need to get query vector in search study.
            query_as_relevant: Array of query row indexes which are in relevant, i.e belong to queries and
                gallery simultaneously.

        Raises:
            ValueError: If dataset has query vector without relevants.
        """
        is_query = query_idxs >= 0
        # array of columns indexes in scores matrix, need to get relevant
        query_column_idxs = query_idxs[is_query]
        # array of row indexes in scores matrix, also this index belong to vectors storage, need to get query vector in
        # search study
        query_row_idxs = np.where(is_query)[0]
        # gallery idxs in vectors storage, which row sum > 0
        # TODO: try to get gallery indexes from Dataset
        gallery_idxs = np.where(np.any(scores > 0, axis=-1))[0]
        # found query row indexes which are in relevant, i.e belong to queries and gallery simultaneously
        query_as_relevant = np.in1d(query_row_idxs, gallery_idxs)

        relevant_idxs = []
        for query_col_idx in query_column_idxs:
            curr_relevant_idxs = np.where(scores[:, query_col_idx] > 0.)[0]
            if len(curr_relevant_idxs) == 0:
                raise ValueError('Representation metric. The dataset contains a query vector that does not '
                                 'has relevants.')
            # Need to sort relevant indexes by its scores for NDCG metric
            current_scores = scores[curr_relevant_idxs, query_col_idx]
            sort_indexes = np.argsort(current_scores)
            curr_relevant_idxs = curr_relevant_idxs[sort_indexes[::-1]]
            relevant_idxs.append(curr_relevant_idxs)

        relevant_idxs = np.array(relevant_idxs)
        return relevant_idxs, gallery_idxs, query_column_idxs, query_row_idxs, query_as_relevant

    def prepare_classification_data(self,
                                    targets: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Prepare data for faiss build index, and following search, in classification dataset case.

        In the classification case, ll the vectors would be used as queries, and the relevants vectors will be vectors
        that have the same label as the query.

        Args:
            targets: Targets in classification task for every vector.

        Returns:
            relevant_idxs: Array of relevant indexes in gallery data, for every query.
            gallery_idxs: Array of gallery indexes in vectors storage.
            query_row_idxs: Array of queries indexes in vectors storage.
            query_as_relevant: Array of query row indexes which are in relevant, i.e belong to queries and
                gallery simultaneously.
        Raises:
            ValueError: If any class has only one element.
        """
        ts = pd.Series(targets)
        groups = pd.Series(ts.groupby(ts).groups)
        # now groups contain group indexes with one targets
        relevant_idxs = []
        query_row_idxs = []
        # create queries with its relevants indexes
        for i, group in enumerate(groups):
            for query_idx in group:
                # need to drop from relevants index which equal query index
                relevant = group.drop(query_idx).tolist()
                if len(relevant) == 0:
                    raise ValueError(f'Representation metric. The class {groups.index[i]} has only one element.')
                query_row_idxs.append(query_idx)
                relevant_idxs.append(relevant)

        relevant_idxs = np.array(relevant_idxs)
        query_row_idxs = np.array(query_row_idxs)
        gallery_idxs = np.arange(len(targets))
        # all query vectors is in relevant, so create array with True elements
        query_as_relevant = np.full((len(gallery_idxs),), fill_value=True, dtype=np.bool)
        return relevant_idxs, gallery_idxs, query_row_idxs, query_as_relevant

    def query_generator(self, index: Union[faiss.IndexFlatIP, faiss.IndexFlatL2],
                        vectors: np.ndarray, relevants_idxs: np.ndarray,
                        query_row_idxs: np.ndarray, gallery_idxs: np.ndarray, query_as_relevant: np.ndarray,
                        k: int, scores: Optional[np.ndarray] = None, query_col_idxs: Optional[np.ndarray] = None
                        ) -> Generator[Tuple[List[np.ndarray], List[np.ndarray]], None, None]:
        """Create relevants and closest arrays, by faiss index search.

        Output in relevant array, contain it's index in gallery data and score for current query.
        Output in closest array, contain it's index in gallery data and distance = 1 for current query.

        This function use self.search_batch_size to define how many vectors to send per one faiss search request.

        Need to know, that query_row_idxs, query_col_idxs and query_as_relevant - have the same size, and for i-th
        query element the following information is available:
            query_row_idxs[i] - index to get embedding vector in vectors storage
            query_col_idxs[i] - column index in scores matrix, need to get score value for every relevant (used in NDCG)
            query_as_relevant[i] - whether the query in relevant i.e in gallery data, need to remove first element in
                retrieved indexes if it true, and last element if it false (because in fact, k + 1 search request is
                being made)

        Args:
            index: Faiss database built index.
            vectors: All embedding vectors, it contains gallery and queries vectors.
            relevants_idxs: Array of relevant indexes for every query, index belong vectors storage.
            query_row_idxs: Array of query indexes in vectors storage.
            gallery_idxs: Array of gallery indexes in vectors storage.
            query_as_relevant: Boolean array of indexes which indicates if query is in relevant set,
                i.e belong to queries and gallery simultaneously.
            k:  Number of top closest indexes to get.
            scores: Array of scores.
            query_col_idxs: Array of query row indexes which are in relevant, i.e belong to queries and
                gallery simultaneously.

        Returns:
            Generator which contain relevant and closest Tuple values.

            Relevant include relevant indexes and scores, size (search_batch_size, , 2).
            Closest include retrieved indexes and scores, size (search_batch_size, , 2).
        """
        for i in range(0, len(query_row_idxs), self.search_batch_size):
            batch_idxs = np.arange(i, min(i + self.search_batch_size, len(query_row_idxs)))

            batch_query_as_relevant = query_as_relevant[batch_idxs]
            batch_query_row_idxs = query_row_idxs[batch_idxs]
            batch_relevants_idxs = relevants_idxs[batch_idxs]

            # queries_idxs - indexes of vectors (global indexes), so queries vectors can be obtained
            # like vectors[queries_idxs[batch_idxs]]
            _, local_closest_idxs = index.search(vectors[batch_query_row_idxs], k=k)
            # get global indexes
            batch_closest_idxs = gallery_idxs[local_closest_idxs]
            # need delete elements same as query elements in searched indexes
            closest_idxs_delete_firs = np.delete(batch_closest_idxs[batch_query_as_relevant], 0, axis=1)
            # or delete last element because we use k + 1 for search
            closest_delete_last = np.delete(batch_closest_idxs[~batch_query_as_relevant], -1, axis=1)
            # create new matrix from closest_idxs_delete_firs and closest_delete_last
            batch_closest_idxs = np.zeros((len(batch_idxs), k - 1))
            batch_closest_idxs[batch_query_as_relevant] = closest_idxs_delete_firs
            batch_closest_idxs[~batch_query_as_relevant] = closest_delete_last

            # NDCG score=distance is needed to sort more relevant examples, but in this part of code we had
            # already sorted our examples by faiss. So if we change score = 1 to distance with type float
            # the index of relevant will be also float and after that inside ranx it may be fail to compare
            # relevant int index with our relevant float index.
            closest_idxs = map(lambda idx:
                               np.stack((batch_closest_idxs[idx], [1] * len(batch_closest_idxs[idx])), axis=1),
                               np.arange(len(batch_closest_idxs)))

            if scores is None or query_col_idxs is None:
                search_relevants_idxs = map(lambda r: np.stack((r, np.ones_like(r)), axis=1), batch_relevants_idxs)
            else:
                batch_query_col_idxs = query_col_idxs[batch_idxs]
                # clear_query_order_numbers cleared of queries, so it has local indexes and the elements can be
                # obtained like clear_query_order_numbers[batch_idxs]
                search_relevants_idxs = map(lambda r_q:
                                            np.stack((r_q[0], scores[r_q[0], r_q[1]]), axis=1),
                                            zip(batch_relevants_idxs, batch_query_col_idxs))

            search_relevants_idxs = list(search_relevants_idxs)
            closest_idxs = list(closest_idxs)
            yield search_relevants_idxs, closest_idxs

    def build_index(self, vectors: np.ndarray):
        """Build index of a given set of vectors with FAISS.

        Args:
            vectors: Database vectors to index, size (database_size, embedding_size).

        Returns:
            Constructed index.
        """
        vectors = vectors.astype(np.float32)
        n, d = vectors.shape

        index_class = faiss.IndexFlatIP if self.metric_distance == MetricDistance.IP else faiss.IndexFlatL2
        if self.exact_index:
            index = index_class(d)
        else:
            nlist = 4 * math.ceil(n ** 0.5)
            quantiser = index_class(d)
            index = faiss.IndexIVFFlat(quantiser, d, nlist, self.metric_distance.value)
            index.train(vectors)

        index.add(vectors)
        return index


@METRICS.register_class
class HitAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True,
                 metric_distance: str = 'IP', k: Optional[int] = None,
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance,
                         metric_func=hit_rate, k=k, search_batch_size=search_batch_size,
                         normalize_vectors=normalize_vectors, **kwargs)


@METRICS.register_class
class PrecisionAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True,
                 metric_distance: str = 'IP', k: Optional[int] = None,
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance,
                         metric_func=precision, k=k, search_batch_size=search_batch_size,
                         normalize_vectors=normalize_vectors, **kwargs)


@METRICS.register_class
class RecallAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True,
                 metric_distance: str = 'IP', k: Optional[int] = None,
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance,
                         metric_func=recall, k=k, search_batch_size=search_batch_size,
                         normalize_vectors=normalize_vectors, **kwargs)


@METRICS.register_class
class MeanAveragePrecisionAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True,
                 metric_distance: str = 'IP', k: Optional[int] = None,
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance,
                         metric_func=average_precision, k=k, search_batch_size=search_batch_size,
                         normalize_vectors=normalize_vectors, **kwargs)


@METRICS.register_class
class NDCGAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True,
                 metric_distance: str = 'IP', k: Optional[int] = None,
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance,
                         metric_func=ndcg, k=k, search_batch_size=search_batch_size,
                         normalize_vectors=normalize_vectors, **kwargs)



from pathlib import Path
from typing import Any, Union, Optional, Dict

import cv2
import torch
import numpy as np
import pandas as pd
from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose

from torchok.data.datasets.base import ImageDataset


class ImageSegmentationDataset(ImageDataset):
    """A dataset for image segmentation task.

    .. csv-table:: Segmentation csv example.
        :header: image_path, mask
        
        image1.png, mask1.png
        image2.png, mask2.png
        image3.png, mask3.png
    """
    def __init__(self,
                 data_folder: str,
                 csv_path: str,
                 transform: Optional[Union[BasicTransform, BaseCompose]],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 image_dtype: str = 'float32',
                 target_dtype: str = 'int64',
                 csv_columns_mapping: Dict[str, str] = None,
                 grayscale: bool = False,
                 test_mode: bool = False):
        """Init ImageSegmentationDataset.

        Args:
            data_folder: Directory with all the images.
            csv_path: Path to the csv file with path to images and masks.
                Path to images must be under column `image_path` and annotations must be under `mask` column.
                User can change column names, if the `csv_columns_mapping` is given.
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations` library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations` library.
            image_dtype: Data type of of the torch tensors related to the image.
            target_dtype: Data type of of the torch tensors related to the target.
            csv_columns_mapping: Matches mapping column names. Key - TorchOK column name, Value - csv column name.
                default value: {'image_path': 'image_path', 'target': 'mask'}
            grayscale: If True, image will be read as grayscale otherwise as RGB.
            test_mode: If True, only image without labels will be returned.
        """
        super().__init__(transform, augment, image_dtype, grayscale, test_mode)

        self.__data_folder = Path(data_folder)
        self.__target_dtype = target_dtype
        self.__csv_path = csv_path
        self.__csv_columns_mapping = csv_columns_mapping if csv_columns_mapping is not None\
            else {'image_path': 'image_path',
                  'target': 'mask'}

        self.__input_column = self.__csv_columns_mapping['image_path']
        self.__target_column = self.__csv_columns_mapping['target']

        self.__csv = pd.read_csv(self.__data_folder / self.__csv_path, dtype={self.__input_column: 'str',
                                                                              self.__target_column: 'str'})

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        record = self.__csv.iloc[idx]
        image_path = self.__data_folder / record[self.__input_column]
        image = self._read_image(image_path)
        sample = {'image': image}

        if not self._test_mode:
            mask_path = self.__data_folder / record[self.__target_column]
            mask = self.__read_mask(mask_path)
            sample['mask'] = mask

        sample = self._apply_transform(self.augment, sample)
        sample = self._apply_transform(self.transform, sample)

        sample['image'] = sample['image'].type(torch.__dict__[self._image_dtype])
        sample['index'] = idx

        if not self._test_mode:
            sample['target'] = sample['mask'].type(torch.__dict__[self.__target_dtype])
            del sample['mask']

        return sample

    def __read_mask(self, mask_path: str) -> np.ndarray:
        """Read mask.

        Args:
            mask_path: Path to mask.

        Raises:
            ValueError: If mask was not read correctly.
        """
        mask = cv2.imread(str(mask_path), 0)

        if mask is None:
            raise ValueError(f'{mask_path} was not read correctly!')

        return mask

    def __len__(self) -> int:
        """Dataset length."""
        return len(self.__csv)

    @property
    def csv_columns_mapping(self) -> dict:
        """Column name matching."""
        return self.__csv_columns_mapping

    @property
    def target_dtype(self) -> str:
        """It's target type."""
        return self.__target_dtype



from typing import Any, Dict, List, Optional, Tuple

from torch import Tensor
from torch.nn import Module, ModuleList


class JointLoss(Module):
    """Represents wrapper for loss modules.

    This joined loss can be forwarded as a weighted sum of losses and can provide direct access to each loss.
    """

    def __init__(self, losses: List[Module], mappings: List[Dict[str, str]],
                 tags: List[Optional[str]], weights: List[Optional[float]],
                 normalize_weights: bool = True):
        """Init JointLoss.

        Reduction isn't applied, so the included loss modules are responsible for reduction

        Args:
            losses: List of loss modules with their `forward` methods implemented
            mappings: List of mappings, each making relationships from neural network outputs to loss inputs.
            Useful in `Task`s in their `forward_with_gt` methods to support direct passing of tensors
            between a neural network and loss functions
            tags: List of tags applied to each loss from `losses`. Useful when user wants to access individual loss
            by its tag via `JointLoss.__getitem__`. Losses for which tags are `None`, can't be accessed directly and
            `JointLoss.__getitem__` will throw a `ValueError`
            weights: List of scalar weights for loss functions, so that resulting loss value will be:
            `loss = loss_i * weight_i for i in [0, len(losses) - 1]`. The weights must be
            either specified for each loss function or not specified for any loss functions.
            In the last case the weights will be automatically set as equal
            normalize_weights: Either to normalize the weights, so that they sum up to 1 before loss value calculation.

        Raises:
            - ValueError: When only a few weights are specified but not for all the loss modules
        """
        super().__init__()
        self.__losses = ModuleList(losses)
        self.__tag2loss = {tag: loss for tag, loss in zip(tags, self.__losses) if tag is not None}
        self.__tags = tags
        self.__mappings = mappings

        num_specified_weights = len(list(filter(lambda w: w is not None, weights)))
        if num_specified_weights > 0 and num_specified_weights != len(losses):
            raise ValueError('Loss weights must be either specified for each loss function or '
                             'not specified for any loss function')

        if num_specified_weights == 0:
            self.__weights = [1.] * len(self.__losses)
        else:
            self.__weights = weights

        if normalize_weights:
            self.__weights = [w / sum(self.__weights) for w in self.__weights]

    def forward(self, **kwargs) -> Tuple[Tensor, Dict[str, Tensor]]:
        """Forward the joined loss module.

        First, individual loss modules are calculated. Then the values are summed up in a weighted average manner
        to form the total loss value.

        Args:
            **kwargs: Any tensors which are supported by the corresponding loss modules.
            The mapping of each loss module is used to map neural network outputs to loss inputs

        Returns:
            Tuple:
            - Tensor, representing total loss value as a weighted sum of all calculated loss values
            - Dict, where key - tag of individual loss module, value - Tensor, representing the loss value.
            Individual values will be return for tagged values only

        Raises:
            - ValueError: When some keys from a mapping cannot be found among neural network outputs
        """
        total_loss = 0.
        tagged_loss_values = {}
        for loss_module, mapping, tag, weight in zip(self.__losses, self.__mappings, self.__tags, self.__weights):
            targeted_kwargs = self.__map_arguments(mapping, **kwargs)
            loss = loss_module(**targeted_kwargs)
            total_loss = total_loss + loss * weight
            if tag is not None:
                tagged_loss_values[tag] = loss

        return total_loss, tagged_loss_values

    def __getitem__(self, tag: str) -> Module:
        """Provide direct access to loss module by its tag.

        Args:
            tag: Tag of the desired loss module

        Returns: Loss module

        Raises:
            - KeyError: When a specified tag cannot be found (often happens when tags are specified not for each loss)
        """
        if tag in self.__tag2loss:
            return self.__tag2loss[tag]
        else:
            raise KeyError(f'Cannot access loss {tag}. You should tag your losses for direct access with a tag key')

    @staticmethod
    def __map_arguments(mapping: Dict[str, str], **model_outputs) -> Dict[str, Any]:
        targeted_kwargs = {}
        for target_arg, source_arg in mapping.items():
            if source_arg in model_outputs:
                targeted_kwargs[target_arg] = model_outputs[source_arg]
            else:
                raise ValueError(f'Cannot find {source_arg} for your mapping {target_arg} : {source_arg}. '
                                 f'You should either add {source_arg} output to your model or remove the mapping '
                                 f'from configuration')
        return targeted_kwargs



import cv2
import hydra
from omegaconf import OmegaConf, DictConfig

from torchok.constructor.config_structure import ConfigParams
from torchok.constructor.runner import create_trainer
from torchok.constructor import TASKS


# Hack to fix multiprocessing deadlock when PyTorch's DataLoader is used
# (more info: https://github.com/pytorch/pytorch/issues/1355)
cv2.setNumThreads(0)


@hydra.main(version_base=None, config_path=None, config_name=None)
def entrypoint(config: DictConfig):
    # Need to add --config-path (-cp) and --config_name (-cn) in run command
    # Example config: examples/configs/classification_cifar10.yaml
    # Then the command will be:
    # python --config-path configs --config_name classification_cifar10

    # Resolve -> change evn variable to values for example ${oc.env:USER} -> 'root'
    OmegaConf.resolve(config)
    # Get entrypoint name - default is train
    entrypoint = config.get('entrypoint', 'train')
    # Remove entrypoint key, because it isn't in ConfigParams
    if 'entrypoint' in config:
        config = dict(config)
        config.pop('entrypoint')
    # Register structure
    schema = OmegaConf.structured(ConfigParams)
    # Merge structure with config
    config = OmegaConf.merge(schema, config)
    # Create task
    model = TASKS.get(config.task.name)(config)
    trainer = create_trainer(config)
    if entrypoint == 'train':
        trainer.fit(model, ckpt_path=config.resume_path)
    elif entrypoint == 'test':
        trainer.test(model, ckpt_path=config.resume_path)
    elif entrypoint == 'predict':
        trainer.predict(model, ckpt_path=config.resume_path)
    else:
        raise ValueError(f'Main function error. Entrypoint with name <{entrypoint}> does not support, please use '
                         f'the following entrypoints - [train, test, predict].')


if __name__ == '__main__':
    entrypoint()



import logging
from typing import Dict, Union, Any

import onnx
import torch
from torch import Tensor
import onnxruntime as onnxrt
from omegaconf import DictConfig

from torchok.constructor import TASKS
from torchok.tasks.base import BaseTask
from torchok.constructor.config_structure import Phase


@TASKS.register_class
class ONNXTask(BaseTask):
    """A class for onnx task."""

    str_type2numpy_type = {'tensor(float)': 'float32', 'tensor(float16)': 'float16',
                           'tensor(double)': 'float64', 'tensor(int8)': 'int8',
                           'tensor(int16)': 'int16', 'tensor(int32)': 'int32',
                           'tensor(int64)': 'int64', 'tensor(uint8)': 'uint8'}

    def __init__(self, hparams: DictConfig):
        """Init ONNXTask.

        Args:
            hparams: Hyperparameters that set in yaml file.
        """
        super().__init__(hparams)
        self.__infer_params = self._hparams.task.params
        model_path = self.__infer_params.path_to_onnx

        onnx_model = onnx.load(model_path)
        onnx.checker.check_model(onnx_model)

        self._sess = onnxrt.InferenceSession(model_path, providers=self.__infer_params.providers)
        self.__binding = self._sess.io_binding()

        self.__inputs = [{'name': item.name,
                          'dtype': self.str_type2numpy_type[item.type]} for item in self._sess.get_inputs()]

        input_names = [input['name'] for input in self.__inputs]
        logging.info(f'ONNX model input names: {input_names}')

        self.__keys_mapping_onnx2dataset = self._hparams.task.params.keys_mapping_onnx2dataset

        self.__outputs = [{'name': item.name,
                           'shape': item.shape,
                           'dtype': self.str_type2numpy_type[item.type]} for item in self._sess.get_outputs()]

        output_names = [output['name'] for output in self.__outputs]
        logging.info(f'ONNX model output names: {output_names}')

    def forward(self, x: Tensor) -> Tensor:
        pass

    def forward_with_gt(self, batch: Dict[str, Any]) -> Dict[str, Tensor]:
        pass

    def foward_infer(self, inputs: Dict[str, Tensor]) -> Dict[str, Tensor]:
        """Forward onnx model."""

        for input in self.__inputs:
            # TODO: Hardcode device_id check that it doesn't matter
            input_tensor = inputs[self.__keys_mapping_onnx2dataset[input['name']]]
            self.batch_dim = input_tensor.shape[0]

            self.__binding.bind_input(
                name=input['name'],
                device_type=self.device,
                device_id=0,
                element_type=input['dtype'],
                shape=input_tensor.shape,
                buffer_ptr=input_tensor.data_ptr())

        output = dict()

        for output_params in self.__outputs:
            output_tensor = torch.empty((self.batch_dim, *output_params['shape'][1:]),
                                        dtype=torch.__dict__[output_params['dtype']],
                                        device=self.device).contiguous()
            self.__binding.bind_output(
                name=output_params['name'],
                device_type=self.device,
                device_id=0,
                element_type=output_params['dtype'],
                shape=(self.batch_dim, *output_params['shape'][1:]),
                buffer_ptr=output_tensor.data_ptr())

            output[output_params['name']] = output_tensor

        self._sess.run_with_iobinding(self.__binding)
        return output

    def forward_infer_with_gt(self, batch: Dict[str, Any]) -> Dict[str, Tensor]:
        """Forward method for test stage."""
        target = batch['target']
        prediction = self.foward_infer(batch)
        output = {'target': target, **prediction}
        return output

    def test_step(self, batch: Dict[str, Union[Tensor, int]], batch_idx: int) -> None:
        """Complete test loop."""
        output = self.forward_infer_with_gt(batch)
        self._metrics_manager.forward(Phase.TEST, **output)

    def predict_step(self, batch: Dict[str, Any], batch_idx: int) -> Tensor:
        """Complete predict loop."""
        output = self.foward_infer(batch)
        return output



import logging
import torch
import pytorch_lightning as pl
from pytorch_lightning.utilities.cloud_io import load
from typing import Optional, Dict, List, Callable, OrderedDict, Union
from collections import defaultdict


def load_state_dict(checkpoint_path: str, map_location: Optional[Union[str, Callable, torch.device]] = 'cpu'):
    """Loads a checkpoint state_dict.

    Args:
        checkpoint_path: Path or URL of the checkpoint, it can be S3 on AWS, GCS on Google Cloud, or ADL on Azure.
        map_location: How to remap storage locations.

    Returns:
        state_dict: Downloaded state dict.
    """
    checkpoint = load(checkpoint_path, map_location=map_location)
    if 'state_dict' in checkpoint:
        state_dict = checkpoint['state_dict']
    else:
        state_dict = checkpoint
    return state_dict


def sort_state_dict_by_depth(override_name2state_dict: Dict[str, str]) -> List[List[OrderedDict[str, torch.Tensor]]]:
    """Generate sorted by depth list of state dict list with the current depth.
    Where depth is calculated as the number of dots in the dictionary key.

    Args:
        override_name2state_dict: Dicts of module key to state dict, which should override base checkpoint.

    Returns:
        depth2override_state_dicts: Sorted by depth dict, where key - depth, value - list of all state dicts with
            current depth.
    """
    depth2override_state_dicts = defaultdict(list)

    for override_key, override_state_dict in override_name2state_dict.items():
        depth = len(override_key.split('.')) - 1
        depth2override_state_dicts[depth].append(override_state_dict)
    # Sort depth2override_state_dicts by it key - depth
    depth2override_state_dicts = OrderedDict(sorted(depth2override_state_dicts.items()))
    return depth2override_state_dicts


def get_state_dict_with_prefix(prefix: str,
                               state_dict: OrderedDict[str, torch.Tensor]) -> OrderedDict[str, torch.Tensor]:
    """Generate state dict with prefixed keys. If input state dict keys startswith prefix then no prefix added.

    Args:
        prefix: Prefix for state dict keys that should be added, then state dict keys not begin with prefix.
        state_dict: State dict whose keys should be prefixed.

    Returns:
        state_dict_with_prefix: Prefixed state dict.
    """
    state_dict_with_prefix = OrderedDict()
    # Remove spaces and dots
    prefix = prefix.strip(' .')
    prefix = prefix + '.'
    for key, value in state_dict.items():
        if not key.startswith(prefix):
            key = prefix + key
        state_dict_with_prefix[key] = value
    return state_dict_with_prefix


def get_absolute_keys(require_key: str, model_keys: List[str]) -> List[str]:
    """Create absolute model keys from require.

    The keys belong to the model state dict called absolute.

    Args:
        require_key: The key to being straightened.
        model_keys: The model keys.

    Returns:
        absolute_keys: Absolute keys.
    """
    absolute_keys = []
    for key in model_keys:
        if key.startswith(require_key):
            absolute_keys.append(key)
    return absolute_keys


def generate_required_state_dict(base_state_dict: OrderedDict[str, torch.Tensor],
                                 overridden_name2state_dict: Dict[str, OrderedDict[str, torch.Tensor]],
                                 exclude_keys: List[str],
                                 model_keys: List[str],
                                 initial_state_dict: OrderedDict[str, torch.Tensor]
                                 ) -> OrderedDict[str, torch.Tensor]:
    """Generate state dict, which should be loaded from 4 main components: base state dict, overridden state dicts,
    exclude keys and model_keys.

    Base state dict values are overridden with overridden_name2state_dict, where the key in overridden_name2state_dict -
    module name which should be overridden.
    If overridden_name2state_dict has many state dicts for one key in the base state dict, then the state dict is
    selected whose overridden_name module is closer to the key i.e. choose the deepest overridden_name, where depth
    is the number of dots.
    After the base state dict had been overridden, it is necessary to remove keys whose names begin with any
    key in exclude_keys.

    Example:
        >>> model_keys = ['backbone.linear.1', 'backbone.linear.2', 'head.linear.1', 'head.linear.2']
        >>> exclude_keys = ['head.linear.2']
        >>> initial_state_dict = {
        >>>    'backbone.linear.1': torch.tensor(0),
        >>>    'backbone.linear.2': torch.tensor(0),
        >>>    'head.linear.1': torch.tensor(0),
        >>>    'head.linear.2': torch.tensor(0)
        >>> }
        >>> base_state_dict = {
        >>>    'backbone.linear.1': torch.tensor(1),
        >>>    'backbone.linear.2': torch.tensor(1),
        >>>    'head.linear.1': torch.tensor(1),
        >>>    'head.linear.2': torch.tensor(1)
        >>> }
        >>> override_backbone = {
        >>>     'backbone.linear.1': torch.tensor(5),
        >>>     'backbone.linear.2': torch.tensor(3)
        >>> }
        >>> override_backbone_linear_1 = {
        >>>     'backbone.linear.1': torch.tensor(10)
        >>> }
        >>> override_name2state_dict = {
        >>>     'backbone': override_backbone,
        >>>     'backbone.linear.1': override_backbone_linear_1
        >>> }
        >>> generate_required_state_dict(base_state_dict, override_name2state_dict,
        >>>                              exclude_keys, model_keys, initial_state_dict)
        {'backbone.linear.1': tensor(10), 'backbone.linear.2': tensor(3),
         'head.linear.1': tensor(1), 'head.linear.2': tensor(0)}

    Args:
        base_state_dict: Base state dict that should be loaded.
        overridden_name2state_dict: Dicts of module key to state dict, which should override base state dict.
        exclude_keys: Module keys that would be loaded from the initial model state dict.
        model_keys: Model state dict keys.
        model_state_dict: Model initial state dict.

    Returns:
        required_state_dict: State dict obtained by override base state dict by overridden state dict, which not
            contain the keys starting with exclude keys.

    Raises:
        ValueError: If any exclude_key not in model.
    """
    required_state_dict = OrderedDict()

    # Add prefix for every overridden state dict
    overridden_full_name2state_dict = dict()
    for overridden_name, state_dict in overridden_name2state_dict.items():
        prefixed_state_dict = get_state_dict_with_prefix(prefix=overridden_name, state_dict=state_dict)
        overridden_full_name2state_dict[overridden_name] = prefixed_state_dict

    # Get sorted by depth state dict list that must be overridden
    depth2state_dicts = sort_state_dict_by_depth(overridden_full_name2state_dict)

    # Firstly change model state dict by base state dict
    required_state_dict.update(base_state_dict)

    # Then change model state dict with overridden state dicts, in order of it's depths
    for _, depth_state_dicts in depth2state_dicts.items():
        for state_dict in depth_state_dicts:
            required_state_dict.update(state_dict)

    # Create absolute exclude keys
    absolute_exclude_keys = []
    for exclude_key in exclude_keys:
        absolute_keys = get_absolute_keys(exclude_key, model_keys)
        if len(absolute_keys) == 0:
            raise ValueError(f'Load checkpoint. Found exclude key {exclude_key} which not in model_keys.')
        absolute_exclude_keys += absolute_keys

    # Create exclude state dict
    exclude_state_dict = OrderedDict()
    for key in absolute_exclude_keys:
        exclude_state_dict[key] = initial_state_dict[key]

    # Update require state dict with exclude state dict
    required_state_dict.update(exclude_state_dict)

    return required_state_dict


def load_checkpoint(model: pl.LightningModule, base_ckpt_path: Optional[str] = None,
                    overridden_name2ckpt_path: Optional[Dict[str, str]] = None,
                    exclude_keys: Optional[List[str]] = None):
    """Load checkpoint to model.

    Args:
        model: Module to load checkpoint into.
        base_ckpt_path: Base checkpoint path that should be loaded.
        override_name2ckpt_path: Dicts of module key to checkpoint path, which should override base checkpoint.
        exclude_names: Module keys that should not be loaded.
    """
    # If no checkpoints to load
    if base_ckpt_path is None and overridden_name2ckpt_path is None:
        logging.warning('Load checkpoint function. You wrote checkpoint parameters in yaml config without base '
                        'checkpoint path and overridden checkpoint paths!')
        return

    initial_state_dict = model.state_dict()
    model_keys = list(initial_state_dict.keys())

    if exclude_keys is None:
        exclude_keys = list()

    # Load base state dict
    base_state_dict = load_state_dict(base_ckpt_path, map_location='cpu') if base_ckpt_path is not None else dict()

    # Load overridden state dicts
    if overridden_name2ckpt_path is None:
        overridden_name2state_dict = dict()
    else:
        overridden_name2state_dict = {name: load_state_dict(ckpt_path)
                                      for name, ckpt_path in overridden_name2ckpt_path.items()}

    required_state_dict = generate_required_state_dict(base_state_dict, overridden_name2state_dict,
                                                       exclude_keys, model_keys, initial_state_dict)

    model.load_state_dict(required_state_dict, strict=True)



""" MobileNet V3
Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/mobilenetv3.py
A PyTorch impl of MobileNet-V3, compatible with TF weights from official impl.


Paper: Searching for MobileNetV3 - https://arxiv.org/abs/1905.02244

Hacked together by / Copyright 2019, Ross Wightman
"""
from functools import partial

import torch.nn as nn
from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD, IMAGENET_INCEPTION_MEAN, IMAGENET_INCEPTION_STD
from timm.models.efficientnet_blocks import SqueezeExcite
from timm.models.efficientnet_builder import BN_EPS_TF_DEFAULT, decode_arch_def, efficientnet_init_weights, \
    EfficientNetBuilder, resolve_act_layer, resolve_bn_args, round_channels
from timm.models.helpers import build_model_with_cfg
from timm.models.layers import create_conv2d

from torchok.constructor import BACKBONES
from torchok.models.backbones.base_backbone import BaseBackbone


def _cfg(url='', **kwargs):
    return {
        'url': url, 'input_size': (3, 224, 224), 'pool_size': (7, 7),
        'crop_pct': 0.875, 'interpolation': 'bilinear',
        'mean': IMAGENET_DEFAULT_MEAN, 'std': IMAGENET_DEFAULT_STD,
        'first_conv': 'conv_stem', 'classifier': 'classifier',
        **kwargs
    }


default_cfgs = {
    'mobilenetv3_large_075': _cfg(url=''),
    'mobilenetv3_large_100': _cfg(
        interpolation='bicubic',
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/mobilenetv3_large_100_ra-f55367f5.pth'),
    'mobilenetv3_large_100_miil': _cfg(
        interpolation='bilinear', mean=(0, 0, 0), std=(1, 1, 1),
        url='https://miil-public-eu.oss-eu-central-1.aliyuncs.com/model-zoo/ImageNet_21K_P/models/timm/mobilenetv3_large_100_1k_miil_78_0.pth'),
    'mobilenetv3_large_100_miil_in21k': _cfg(
        interpolation='bilinear', mean=(0, 0, 0), std=(1, 1, 1),
        url='https://miil-public-eu.oss-eu-central-1.aliyuncs.com/model-zoo/ImageNet_21K_P/models/timm/mobilenetv3_large_100_in21k_miil.pth',
        num_classes=11221),

    'mobilenetv3_small_050': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/mobilenetv3_small_050_lambc-4b7bbe87.pth',
        interpolation='bicubic'),
    'mobilenetv3_small_075': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/mobilenetv3_small_075_lambc-384766db.pth',
        interpolation='bicubic'),
    'mobilenetv3_small_100': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/mobilenetv3_small_100_lamb-266a294c.pth',
        interpolation='bicubic'),

    'mobilenetv3_rw': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/mobilenetv3_100-35495452.pth',
        interpolation='bicubic'),

    'tf_mobilenetv3_large_075': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/tf_mobilenetv3_large_075-150ee8b0.pth',
        mean=IMAGENET_INCEPTION_MEAN, std=IMAGENET_INCEPTION_STD),
    'tf_mobilenetv3_large_100': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/tf_mobilenetv3_large_100-427764d5.pth',
        mean=IMAGENET_INCEPTION_MEAN, std=IMAGENET_INCEPTION_STD),
    'tf_mobilenetv3_large_minimal_100': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/tf_mobilenetv3_large_minimal_100-8596ae28.pth',
        mean=IMAGENET_INCEPTION_MEAN, std=IMAGENET_INCEPTION_STD),
    'tf_mobilenetv3_small_075': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/tf_mobilenetv3_small_075-da427f52.pth',
        mean=IMAGENET_INCEPTION_MEAN, std=IMAGENET_INCEPTION_STD),
    'tf_mobilenetv3_small_100': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/tf_mobilenetv3_small_100-37f49e2b.pth',
        mean=IMAGENET_INCEPTION_MEAN, std=IMAGENET_INCEPTION_STD),
    'tf_mobilenetv3_small_minimal_100': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/tf_mobilenetv3_small_minimal_100-922a7843.pth',
        mean=IMAGENET_INCEPTION_MEAN, std=IMAGENET_INCEPTION_STD),

    'fbnetv3_b': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/fbnetv3_b_224-ead5d2a1.pth',
        test_input_size=(3, 256, 256), crop_pct=0.95),
    'fbnetv3_d': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/fbnetv3_d_224-c98bce42.pth',
        test_input_size=(3, 256, 256), crop_pct=0.95),
    'fbnetv3_g': _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/fbnetv3_g_240-0b1df83b.pth',
        input_size=(3, 240, 240), test_input_size=(3, 288, 288), crop_pct=0.95, pool_size=(8, 8)),

    "lcnet_035": _cfg(),
    "lcnet_050": _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/lcnet_050-f447553b.pth',
        interpolation='bicubic',
    ),
    "lcnet_075": _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/lcnet_075-318cad2c.pth',
        interpolation='bicubic',
    ),
    "lcnet_100": _cfg(
        url='https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-weights/lcnet_100-a929038c.pth',
        interpolation='bicubic',
    ),
    "lcnet_150": _cfg(),
}


class MobileNetV3(BaseBackbone):
    """ MobiletNet-V3

    Based on my EfficientNet implementation and building blocks, this model utilizes the MobileNet-v3 specific
    'efficient head', where global pooling is done before the head convolution without a final batch-norm
    layer before the classifier.

    Paper: `Searching for MobileNetV3` - https://arxiv.org/abs/1905.02244

    Other architectures utilizing MobileNet-V3 efficient head that are supported by this impl include:
      * HardCoRe-NAS - https://arxiv.org/abs/2102.11646 (defn in hardcorenas.py uses this class)
      * FBNet-V3 - https://arxiv.org/abs/2006.02049
      * LCNet - https://arxiv.org/abs/2109.15099
    """

    def __init__(self, block_args, in_channels=3, stem_size=16, fix_stem=False, num_features=1280, pad_type='',
                 act_layer=None, norm_layer=None, se_layer=None, se_from_exp=True, round_chs_fn=round_channels,
                 output_stride=32, drop_path_rate=0.):
        super(MobileNetV3, self).__init__(in_channels=in_channels, out_channels=num_features)
        act_layer = act_layer or nn.ReLU
        norm_layer = norm_layer or nn.BatchNorm2d
        se_layer = se_layer or SqueezeExcite
        self.num_features = num_features

        # Stem
        if not fix_stem:
            stem_size = round_chs_fn(stem_size)
        self.conv_stem = create_conv2d(in_channels, stem_size, 3, stride=2, padding=pad_type)
        self.bn1 = norm_layer(stem_size)
        self.act1 = act_layer(inplace=True)

        # Middle stages (IR/ER/DS Blocks)
        builder = EfficientNetBuilder(
            output_stride=output_stride, pad_type=pad_type, round_chs_fn=round_chs_fn, se_from_exp=se_from_exp,
            act_layer=act_layer, norm_layer=norm_layer, se_layer=se_layer,
            drop_path_rate=drop_path_rate)
        self.blocks = nn.Sequential(*builder(stem_size, block_args))
        self.feature_info = builder.features

        self.create_hooks()
        efficientnet_init_weights(self)

    def forward(self, x):
        x = self.conv_stem(x)
        x = self.bn1(x)
        x = self.blocks(x)
        return x


def create_mnv3(variant, pretrained=False, **kwargs):
    return build_model_with_cfg(MobileNetV3, variant, pretrained, pretrained_strict=False,
                                kwargs_filter=('num_classes', 'global_pool', 'in_chans', 'head_bias'), **kwargs)


def _gen_mobilenet_v3_rw(variant, channel_multiplier=1.0, pretrained=False, **kwargs):
    """Creates a MobileNet-V3 model.

    Ref impl: ?
    Paper: https://arxiv.org/abs/1905.02244

    Args:
      channel_multiplier: multiplier to number of channels per layer.
    """
    arch_def = [
        # stage 0, 112x112 in
        ['ds_r1_k3_s1_e1_c16_nre_noskip'],  # relu
        # stage 1, 112x112 in
        ['ir_r1_k3_s2_e4_c24_nre', 'ir_r1_k3_s1_e3_c24_nre'],  # relu
        # stage 2, 56x56 in
        ['ir_r3_k5_s2_e3_c40_se0.25_nre'],  # relu
        # stage 3, 28x28 in
        ['ir_r1_k3_s2_e6_c80', 'ir_r1_k3_s1_e2.5_c80', 'ir_r2_k3_s1_e2.3_c80'],  # hard-swish
        # stage 4, 14x14in
        ['ir_r2_k3_s1_e6_c112_se0.25'],  # hard-swish
        # stage 5, 14x14in
        ['ir_r3_k5_s2_e6_c160_se0.25'],  # hard-swish
        # stage 6, 7x7 in
        ['cn_r1_k1_s1_c960'],  # hard-swish
    ]
    model_kwargs = dict(
        block_args=decode_arch_def(arch_def),
        head_bias=False,
        round_chs_fn=partial(round_channels, multiplier=channel_multiplier),
        norm_layer=partial(nn.BatchNorm2d, **resolve_bn_args(kwargs)),
        act_layer=resolve_act_layer(kwargs, 'hard_swish'),
        se_layer=partial(SqueezeExcite, gate_layer='hard_sigmoid'),
        **kwargs,
    )
    model = create_mnv3(variant, pretrained, **model_kwargs)
    return model


def _gen_mobilenet_v3(variant, channel_multiplier=1.0, pretrained=False, **kwargs):
    """Creates a MobileNet-V3 model.

    Ref impl: ?
    Paper: https://arxiv.org/abs/1905.02244

    Args:
      channel_multiplier: multiplier to number of channels per layer.
    """
    if 'small' in variant:
        num_features = 1024
        if 'minimal' in variant:
            act_layer = resolve_act_layer(kwargs, 'relu')
            arch_def = [
                # stage 0, 112x112 in
                ['ds_r1_k3_s2_e1_c16'],
                # stage 1, 56x56 in
                ['ir_r1_k3_s2_e4.5_c24', 'ir_r1_k3_s1_e3.67_c24'],
                # stage 2, 28x28 in
                ['ir_r1_k3_s2_e4_c40', 'ir_r2_k3_s1_e6_c40'],
                # stage 3, 14x14 in
                ['ir_r2_k3_s1_e3_c48'],
                # stage 4, 14x14in
                ['ir_r3_k3_s2_e6_c96'],
                # stage 6, 7x7 in
                ['cn_r1_k1_s1_c576'],
            ]
        else:
            act_layer = resolve_act_layer(kwargs, 'hard_swish')
            arch_def = [
                # stage 0, 112x112 in
                ['ds_r1_k3_s2_e1_c16_se0.25_nre'],  # relu
                # stage 1, 56x56 in
                ['ir_r1_k3_s2_e4.5_c24_nre', 'ir_r1_k3_s1_e3.67_c24_nre'],  # relu
                # stage 2, 28x28 in
                ['ir_r1_k5_s2_e4_c40_se0.25', 'ir_r2_k5_s1_e6_c40_se0.25'],  # hard-swish
                # stage 3, 14x14 in
                ['ir_r2_k5_s1_e3_c48_se0.25'],  # hard-swish
                # stage 4, 14x14in
                ['ir_r3_k5_s2_e6_c96_se0.25'],  # hard-swish
                # stage 6, 7x7 in
                ['cn_r1_k1_s1_c576'],  # hard-swish
            ]
    else:
        num_features = 1280
        if 'minimal' in variant:
            act_layer = resolve_act_layer(kwargs, 'relu')
            arch_def = [
                # stage 0, 112x112 in
                ['ds_r1_k3_s1_e1_c16'],
                # stage 1, 112x112 in
                ['ir_r1_k3_s2_e4_c24', 'ir_r1_k3_s1_e3_c24'],
                # stage 2, 56x56 in
                ['ir_r3_k3_s2_e3_c40'],
                # stage 3, 28x28 in
                ['ir_r1_k3_s2_e6_c80', 'ir_r1_k3_s1_e2.5_c80', 'ir_r2_k3_s1_e2.3_c80'],
                # stage 4, 14x14in
                ['ir_r2_k3_s1_e6_c112'],
                # stage 5, 14x14in
                ['ir_r3_k3_s2_e6_c160'],
                # stage 6, 7x7 in
                ['cn_r1_k1_s1_c960'],
            ]
        else:
            act_layer = resolve_act_layer(kwargs, 'hard_swish')
            arch_def = [
                # stage 0, 112x112 in
                ['ds_r1_k3_s1_e1_c16_nre'],  # relu
                # stage 1, 112x112 in
                ['ir_r1_k3_s2_e4_c24_nre', 'ir_r1_k3_s1_e3_c24_nre'],  # relu
                # stage 2, 56x56 in
                ['ir_r3_k5_s2_e3_c40_se0.25_nre'],  # relu
                # stage 3, 28x28 in
                ['ir_r1_k3_s2_e6_c80', 'ir_r1_k3_s1_e2.5_c80', 'ir_r2_k3_s1_e2.3_c80'],  # hard-swish
                # stage 4, 14x14in
                ['ir_r2_k3_s1_e6_c112_se0.25'],  # hard-swish
                # stage 5, 14x14in
                ['ir_r3_k5_s2_e6_c160_se0.25'],  # hard-swish
                # stage 6, 7x7 in
                ['cn_r1_k1_s1_c960'],  # hard-swish
            ]
    se_layer = partial(SqueezeExcite, gate_layer='hard_sigmoid', force_act_layer=nn.ReLU, rd_round_fn=round_channels)
    model_kwargs = dict(
        block_args=decode_arch_def(arch_def),
        num_features=num_features,
        stem_size=16,
        fix_stem=channel_multiplier < 0.75,
        round_chs_fn=partial(round_channels, multiplier=channel_multiplier),
        norm_layer=partial(nn.BatchNorm2d, **resolve_bn_args(kwargs)),
        act_layer=act_layer,
        se_layer=se_layer,
        **kwargs,
    )
    model = create_mnv3(variant, pretrained, **model_kwargs)
    return model


def _gen_fbnetv3(variant, channel_multiplier=1.0, pretrained=False, **kwargs):
    """ FBNetV3
    Paper: `FBNetV3: Joint Architecture-Recipe Search using Predictor Pretraining`
        - https://arxiv.org/abs/2006.02049
    FIXME untested, this is a preliminary impl of some FBNet-V3 variants.
    """
    vl = variant.split('_')[-1]
    if vl in ('a', 'b'):
        stem_size = 16
        arch_def = [
            ['ds_r2_k3_s1_e1_c16'],
            ['ir_r1_k5_s2_e4_c24', 'ir_r3_k5_s1_e2_c24'],
            ['ir_r1_k5_s2_e5_c40_se0.25', 'ir_r4_k5_s1_e3_c40_se0.25'],
            ['ir_r1_k5_s2_e5_c72', 'ir_r4_k3_s1_e3_c72'],
            ['ir_r1_k3_s1_e5_c120_se0.25', 'ir_r5_k5_s1_e3_c120_se0.25'],
            ['ir_r1_k3_s2_e6_c184_se0.25', 'ir_r5_k5_s1_e4_c184_se0.25', 'ir_r1_k5_s1_e6_c224_se0.25'],
            ['cn_r1_k1_s1_c1344'],
        ]
    elif vl == 'd':
        stem_size = 24
        arch_def = [
            ['ds_r2_k3_s1_e1_c16'],
            ['ir_r1_k3_s2_e5_c24', 'ir_r5_k3_s1_e2_c24'],
            ['ir_r1_k5_s2_e4_c40_se0.25', 'ir_r4_k3_s1_e3_c40_se0.25'],
            ['ir_r1_k3_s2_e5_c72', 'ir_r4_k3_s1_e3_c72'],
            ['ir_r1_k3_s1_e5_c128_se0.25', 'ir_r6_k5_s1_e3_c128_se0.25'],
            ['ir_r1_k3_s2_e6_c208_se0.25', 'ir_r5_k5_s1_e5_c208_se0.25', 'ir_r1_k5_s1_e6_c240_se0.25'],
            ['cn_r1_k1_s1_c1440'],
        ]
    elif vl == 'g':
        stem_size = 32
        arch_def = [
            ['ds_r3_k3_s1_e1_c24'],
            ['ir_r1_k5_s2_e4_c40', 'ir_r4_k5_s1_e2_c40'],
            ['ir_r1_k5_s2_e4_c56_se0.25', 'ir_r4_k5_s1_e3_c56_se0.25'],
            ['ir_r1_k5_s2_e5_c104', 'ir_r4_k3_s1_e3_c104'],
            ['ir_r1_k3_s1_e5_c160_se0.25', 'ir_r8_k5_s1_e3_c160_se0.25'],
            ['ir_r1_k3_s2_e6_c264_se0.25', 'ir_r6_k5_s1_e5_c264_se0.25', 'ir_r2_k5_s1_e6_c288_se0.25'],
            ['cn_r1_k1_s1_c1728'],
        ]
    else:
        raise NotImplemented
    round_chs_fn = partial(round_channels, multiplier=channel_multiplier, round_limit=0.95)
    se_layer = partial(SqueezeExcite, gate_layer='hard_sigmoid', rd_round_fn=round_chs_fn)
    act_layer = resolve_act_layer(kwargs, 'hard_swish')
    model_kwargs = dict(
        block_args=decode_arch_def(arch_def),
        num_features=1984,
        head_bias=False,
        stem_size=stem_size,
        round_chs_fn=round_chs_fn,
        se_from_exp=False,
        norm_layer=partial(nn.BatchNorm2d, **resolve_bn_args(kwargs)),
        act_layer=act_layer,
        se_layer=se_layer,
        **kwargs,
    )
    model = create_mnv3(variant, pretrained, **model_kwargs)
    return model


def _gen_lcnet(variant, channel_multiplier=1.0, pretrained=False, **kwargs):
    """ LCNet
    Essentially a MobileNet-V3 crossed with a MobileNet-V1

    Paper: `PP-LCNet: A Lightweight CPU Convolutional Neural Network` - https://arxiv.org/abs/2109.15099

    Args:
      channel_multiplier: multiplier to number of channels per layer.
    """
    arch_def = [
        # stage 0, 112x112 in
        ['dsa_r1_k3_s1_c32'],
        # stage 1, 112x112 in
        ['dsa_r2_k3_s2_c64'],
        # stage 2, 56x56 in
        ['dsa_r2_k3_s2_c128'],
        # stage 3, 28x28 in
        ['dsa_r1_k3_s2_c256', 'dsa_r1_k5_s1_c256'],
        # stage 4, 14x14in
        ['dsa_r4_k5_s1_c256'],
        # stage 5, 14x14in
        ['dsa_r2_k5_s2_c512_se0.25'],
        # 7x7
    ]
    model_kwargs = dict(
        block_args=decode_arch_def(arch_def),
        stem_size=16,
        round_chs_fn=partial(round_channels, multiplier=channel_multiplier),
        norm_layer=partial(nn.BatchNorm2d, **resolve_bn_args(kwargs)),
        act_layer=resolve_act_layer(kwargs, 'hard_swish'),
        se_layer=partial(SqueezeExcite, gate_layer='hard_sigmoid', force_act_layer=nn.ReLU),
        num_features=1280,
        **kwargs,
    )
    model = create_mnv3(variant, pretrained, **model_kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_large_075(pretrained=False, **kwargs):
    """ MobileNet V3 """
    model = _gen_mobilenet_v3('mobilenetv3_large_075', 0.75, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_large_100(pretrained=False, **kwargs):
    """ MobileNet V3 """
    model = _gen_mobilenet_v3('mobilenetv3_large_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_large_100_miil(pretrained=False, **kwargs):
    """ MobileNet V3
    Weights taken from: https://github.com/Alibaba-MIIL/ImageNet21K
    """
    model = _gen_mobilenet_v3('mobilenetv3_large_100_miil', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_large_100_miil_in21k(pretrained=False, **kwargs):
    """ MobileNet V3, 21k pretraining
    Weights taken from: https://github.com/Alibaba-MIIL/ImageNet21K
    """
    model = _gen_mobilenet_v3('mobilenetv3_large_100_miil_in21k', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_small_050(pretrained=False, **kwargs):
    """ MobileNet V3 """
    model = _gen_mobilenet_v3('mobilenetv3_small_050', 0.50, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_small_075(pretrained=False, **kwargs):
    """ MobileNet V3 """
    model = _gen_mobilenet_v3('mobilenetv3_small_075', 0.75, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_small_100(pretrained=False, **kwargs):
    """ MobileNet V3 """
    model = _gen_mobilenet_v3('mobilenetv3_small_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def mobilenetv3_rw(pretrained=False, **kwargs):
    """ MobileNet V3 """
    if pretrained:
        # pretrained model trained with non-default BN epsilon
        kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    model = _gen_mobilenet_v3_rw('mobilenetv3_rw', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def tf_mobilenetv3_large_075(pretrained=False, **kwargs):
    """ MobileNet V3 """
    kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    kwargs['pad_type'] = 'same'
    model = _gen_mobilenet_v3('tf_mobilenetv3_large_075', 0.75, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def tf_mobilenetv3_large_100(pretrained=False, **kwargs):
    """ MobileNet V3 """
    kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    kwargs['pad_type'] = 'same'
    model = _gen_mobilenet_v3('tf_mobilenetv3_large_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def tf_mobilenetv3_large_minimal_100(pretrained=False, **kwargs):
    """ MobileNet V3 """
    kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    kwargs['pad_type'] = 'same'
    model = _gen_mobilenet_v3('tf_mobilenetv3_large_minimal_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def tf_mobilenetv3_small_075(pretrained=False, **kwargs):
    """ MobileNet V3 """
    kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    kwargs['pad_type'] = 'same'
    model = _gen_mobilenet_v3('tf_mobilenetv3_small_075', 0.75, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def tf_mobilenetv3_small_100(pretrained=False, **kwargs):
    """ MobileNet V3 """
    kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    kwargs['pad_type'] = 'same'
    model = _gen_mobilenet_v3('tf_mobilenetv3_small_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def tf_mobilenetv3_small_minimal_100(pretrained=False, **kwargs):
    """ MobileNet V3 """
    kwargs['bn_eps'] = BN_EPS_TF_DEFAULT
    kwargs['pad_type'] = 'same'
    model = _gen_mobilenet_v3('tf_mobilenetv3_small_minimal_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def fbnetv3_b(pretrained=False, **kwargs):
    """ FBNetV3-B """
    model = _gen_fbnetv3('fbnetv3_b', pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def fbnetv3_d(pretrained=False, **kwargs):
    """ FBNetV3-D """
    model = _gen_fbnetv3('fbnetv3_d', pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def fbnetv3_g(pretrained=False, **kwargs):
    """ FBNetV3-G """
    model = _gen_fbnetv3('fbnetv3_g', pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def lcnet_035(pretrained=False, **kwargs):
    """ PP-LCNet 0.35"""
    model = _gen_lcnet('lcnet_035', 0.35, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def lcnet_050(pretrained=False, **kwargs):
    """ PP-LCNet 0.5"""
    model = _gen_lcnet('lcnet_050', 0.5, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def lcnet_075(pretrained=False, **kwargs):
    """ PP-LCNet 1.0"""
    model = _gen_lcnet('lcnet_075', 0.75, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def lcnet_100(pretrained=False, **kwargs):
    """ PP-LCNet 1.0"""
    model = _gen_lcnet('lcnet_100', 1.0, pretrained=pretrained, **kwargs)
    return model


@BACKBONES.register_class
def lcnet_150(pretrained=False, **kwargs):
    """ PP-LCNet 1.5"""
    model = _gen_lcnet('lcnet_150', 1.5, pretrained=pretrained, **kwargs)
    return model



from pathlib import Path
from typing import Union, Optional
import pandas as pd

import torch
from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose
from torchok.data.datasets.base import ImageDataset
from torchvision.datasets.utils import download_and_extract_archive

from torchok.constructor import DATASETS


@DATASETS.register_class
class SOP(ImageDataset):
    """A class represent Stanford Online Products - SOP dataset.
    
    Additionally, we collected Stanford Online Products dataset: 120k images of 23k classes of online products 
    for metric learning. The homepage of SOP is https://cvgl.stanford.edu/projects/lifted_struct/.
    """
    base_folder = 'Stanford_Online_Products'
    filename = 'Stanford_Online_Products.tar.gz'
    
    url = 'https://torchok-hub.s3.eu-west-1.amazonaws.com/Stanford_Online_Products.tar.gz'
    tgz_md5 = '26513716999698fd361a21c93f77ed32'

    train_txt = 'Ebay_train.txt'
    test_txt = 'Ebay_test.txt'

    def __init__(self,
                 train: bool,
                 download: bool,
                 data_folder: str,
                 transform: Optional[Union[BasicTransform, BaseCompose]],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 image_dtype: str = 'float32',
                 grayscale: bool = False,
                 test_mode: bool = False):
        """Init SOP.

        Have 120,053 images with 22,634 classes in the dataset in total. 
        Train have 59551 images with 11318 classes.
        Test have 60502 images with 11316 classes.

        Args:
            train: If True, train dataset will be used, else - test dataset.
            download: If True, data will be download and save to data_folder.
            data_folder: Directory with all the images.
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations` library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations` library.
            image_dtype: Data type of of the torch tensors related to the image.
            grayscale: If True, image will be read as grayscale otherwise as RGB.
            test_mode: If True, only image without labels will be returned.
        """
        super().__init__(transform, augment, image_dtype, grayscale, test_mode)
        self.__data_folder = Path(data_folder)
        self.__path = self.__data_folder / self.base_folder
        self.__train = train

        if download:
            self.__download()

        if not self.__path.is_dir():
            raise RuntimeError('Dataset not found or corrupted. You can use download=True to download it')

        if self.__train:
            self.__csv = pd.read_csv(self.__path / self.train_txt, sep=' ')
        else:
            self.__csv = pd.read_csv(self.__path / self.test_txt, sep=' ')

        self.__target_column = 'class_id'
        self.__path_column = 'path'

    def __getitem__(self, idx: int) -> dict:
        """Get item sample.

        Returns:
            sample: dict, where
            sample['image'] - Tensor, representing image after augmentations and transformations, dtype=image_dtype.
            sample['target'] - Target class or labels, dtype=target_dtype.
            sample['index'] - Index.
        """
        record = self.__csv.iloc[idx]
        image = self._read_image(self.__path / record[self.__path_column])
        sample = {"image": image}
        sample = self._apply_transform(self.augment, sample)
        sample = self._apply_transform(self.transform, sample)
        sample['image'] = sample['image'].type(torch.__dict__[self._image_dtype])
        sample['index'] = idx

        if self._test_mode:
            return sample

        if self.__train:
            # The labels starts with 1 for train
            sample['target'] = record[self.__target_column] - 1
        else:
            # The labels starts with 11319 for train
            sample['target'] = record[self.__target_column] - 11319

        return sample

    def __len__(self) -> int:
        """Dataset length."""
        return len(self.__csv)

    def __download(self) -> None:
        """Download archive by url to specific folder."""
        if self.__path.is_dir():
            print('Files already downloaded and verified')
        else:
            download_and_extract_archive(self.url, self.__data_folder, filename=self.filename, md5=self.tgz_md5)
